from math import exp

from .utils import Fungicide
from .params import PARAMS



class ODESystem:
    def __init__(self, fungicide_params) -> None:
        if fungicide_params is None:
            omega_1 = PARAMS.omega_1
            omega_2 = PARAMS.omega_2
            
            theta_1 = PARAMS.theta_1
            theta_2 = PARAMS.theta_2

            delta_1 = PARAMS.delta_1
            delta_2 = PARAMS.delta_2

        else:
            omega_1 = fungicide_params['omega_1']
            omega_2 = fungicide_params['omega_2']
            
            theta_1 = fungicide_params['theta_1']
            theta_2 = fungicide_params['theta_2']

            delta_1 = fungicide_params['delta_1']
            delta_2 = fungicide_params['delta_2']

        self.fcide1 = Fungicide(omega_1, theta_1, delta_1)
        self.fcide2 = Fungicide(omega_2, theta_2, delta_2)
    



    def system(self, t, y):

        S,ER,ERS,ESR,ES,IR,IRS,ISR,IS,R,PR,PRS,PSR,PS,conc_1,conc_2 = y

        A = S + ER + ERS + ESR + ES + IR + IRS + ISR + IS + R

        dydt = [self._growth(A,t)
             - (self._senescence(t))*S
             -  S * (PARAMS.beta/A) * (
                  (IR + PR)
                + (IRS + PRS) * (self.fcide2.effect(conc_2))
                + (ISR + PSR) * (self.fcide1.effect(conc_1))
                + (IS  +  PS) * (self.fcide1.effect(conc_1)) * (self.fcide2.effect(conc_2))),
            
            S*(PARAMS.beta/A) * (IR + PR) - (self._senescence(t)) * ER  - PARAMS.gamma * ER,
            S*(PARAMS.beta/A) * (IRS + PRS) * (self.fcide2.effect(conc_2)) - (self._senescence(t)) * ERS - PARAMS.gamma * (self.fcide2.effect(conc_2)) * ERS,
            S*(PARAMS.beta/A) * (ISR + PSR) * (self.fcide1.effect(conc_1)) - (self._senescence(t)) * ESR - PARAMS.gamma * (self.fcide1.effect(conc_1)) * ESR,
            S*(PARAMS.beta/A) * (IS  +  PS) * (self.fcide1.effect(conc_1)) * (self.fcide2.effect(conc_2)) - (self._senescence(t)) * ES  - PARAMS.gamma * (self.fcide1.effect(conc_1))*(self.fcide2.effect(conc_2)) * ES,
            
            PARAMS.gamma * ER   -  PARAMS.mu * IR,
            PARAMS.gamma * (self.fcide2.effect(conc_2)) * ERS  -  PARAMS.mu * IRS,
            PARAMS.gamma * (self.fcide1.effect(conc_1)) * ESR  -  PARAMS.mu * ISR,
            PARAMS.gamma * (self.fcide1.effect(conc_1)) * (self.fcide2.effect(conc_2)) * ES   -  PARAMS.mu * IS,
            
            PARAMS.mu * (IR + IRS + ISR + IS)   +  (self._senescence(t)) * (S + ER + ERS + ESR + ES),
            
            - PARAMS.nu * PR,
            - PARAMS.nu * PRS,
            - PARAMS.nu * PSR,
            - PARAMS.nu * PS,
            
            - self.fcide1.delta * conc_1,
            - self.fcide2.delta * conc_2
            ]

        return dydt


    @staticmethod
    def _growth(A, t):
        if t>=PARAMS.T_emerge:
            grw = PARAMS.r*(PARAMS.k-A)
            return grw
        else:
            return 0


    @staticmethod
    def _senescence(t):
        if t>=PARAMS.T_GS61:
            out = 0.005*((t-PARAMS.T_GS61)/(PARAMS.T_GS87-PARAMS.T_GS61)) + 0.1*exp(-0.02*(PARAMS.T_GS87-t))
            return out
        else:
            return 0







import plotly.colors as pltly_clrs
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import numpy as np
import pandas as pd
from math import log2, floor, log10, pi
from scipy import stats
from PIL import Image



from model.strategy_arrays import EqualResFreqBreakdownArray, EqualSelectionArray

from plotting.traces import get_RFB_diff_traces, get_eq_sel_traces, get_heatmap_lines, \
    get_strain_freq_traces, contour_at_0, get_multi_contour_traces, \
    get_MS_RFB_traces, contour_at_single_level

from plotting.utils import get_text_annotation, get_arrow_annotation, standard_layout, \
    grey_colorscale, my_colorbar, get_big_text_annotation

from plotting.consts import ATTRS_DICT, TITLE_MAP, PLOT_WIDTH, PLOT_HEIGHT, \
        FULL_PAGE_WIDTH



# TOC
# Single Tactic
# Changing dose
# Changing fcide
# Grid of tactics
# Dose space
# RF Ratio
# Paper Figs



#----------------------------------------------------------------------------------------------
# * Single Tactic

def yield_by_year(data, conf_str):
    traces = []
    
    y = data['yield_vec']
    x = list(range(1,1+len(y)))

    line = go.Scatter(
        x = x,
        y = y,
        line = dict(color="green"),
    )

    y_low = y[-1]-2
    
    annotz = [dict(
        xref="x1",
        yref="y1",
        x=10,
        y=0.5*(95+y_low),
        text="Unacceptable yield",
        showarrow=False,
        )]
    
    shape = go.Scatter(x=[0, 0, x[-1], x[-1]],
                        y=[y_low, 95, 95, y_low],
                        fill="toself",
                        mode="lines",
                        showlegend=False,
                        line=dict(width=0, color="rgb(150,150,150)"))
    
    traces.append(line)
    traces.append(shape)

    fig = go.Figure(data=traces, layout=standard_layout(False))
    fig.update_layout(annotations=annotz)

    fig.update_xaxes(title="Year")
    fig.update_yaxes(title="Yield<br>(% of disease free)")

    fig.show()
    filename = conf_str.replace("/single/", "/single/yield_by_year/")
    fig.write_image(filename)




def res_freqs_single_t_plot(data, conf_str):
    traces = []

    titles = dict(
        f1 = "Fungicide A",
        f2 = "Fungicide B"
    )
    
    for key in ['f1', 'f2']:
        y = data['res_vec_dict'][key]
        x = list(range(len(y)))

        line = go.Scatter(
            x = x,
            y = y,
            name = titles[key]
        )

        traces.append(line)

    fig = go.Figure(data=traces, layout=standard_layout(True))

    fig.update_xaxes(title="Year")
    fig.update_yaxes(title="Resistant frequency")

    fig.show()
    filename = conf_str.replace("/single/", "/single/res_freqs/")
    fig.write_image(filename)



def single_year_plot(data, indices, conf_str):
    traces = []
    
    for ind in indices:
        y = data['sol_array'][:, ind, 0]
        x = data['t_vec']

        line = go.Scatter(
            x = x,
            y = y,
            name = ATTRS_DICT[str(ind)]['name'],
            line = dict(color=ATTRS_DICT[str(ind)]['colour'], dash=ATTRS_DICT[str(ind)]['dash'])
        )

        traces.append(line)

    fig = go.Figure(data=traces, layout=standard_layout(True))

    fig.update_xaxes(title="Time (degree-days)")
    fig.update_yaxes(title="Amount")
    
    if 15 in indices:
        fig.update_layout(legend=dict(x=0.5, y=0.85))
    elif 0 in indices:
        fig.update_layout(legend=dict(x=0.2, y=0.55))
    else:
        fig.update_layout(legend=dict(x=0.01, y=0.95))
    
    fig.update_layout(width=PLOT_WIDTH*(2/3), height=PLOT_HEIGHT)

    fig.show()
    filename = conf_str.replace("/single/", "/single/within_season/plot" + "".join(str(e) for e in indices))
    fig.write_image(filename)



def yield_res_freqs_plot(data, conf_str):
    fig = make_subplots(rows=2, cols=1, shared_xaxes=True)
    
    rf_traces = []

    titles = dict(
        f1 = "Fungicide A",
        f2 = "Fungicide B"
    )
    
    for key in ['f1', 'f2']:
        y = data['res_vec_dict'][key]
        x = list(range(len(y)))

        line = go.Scatter(
            x = x,
            y = y,
            mode="lines+markers",
            name = f"Resistance to {titles[key].lower()}"
        )

        rf_traces.append(line)
        fig.add_trace(line, row=2, col=1)

    fig.update_xaxes(title="Year", row=2, col=1, showgrid=False, zeroline=False)
    fig.update_yaxes(title="Resistant<br>frequency", row=2, col=1)


    y = data['yield_vec']
    x = list(range(1,1+len(y)))

    line = go.Scatter(
        x = x,
        y = y,
        line=dict(color="green"),
        mode="lines+markers",
        name="Yield",
        showlegend=False,
    )

    y_low = y[-1]-2
    
    annotz = [dict(
        xref="x1",
        yref="y1",
        x=5,
        y=0.5*(95+y_low),
        text="Unacceptable yield",
        showarrow=False,
        )]

    shape = go.Scatter(x=[0, 0, x[-1], x[-1]],
                        y=[y_low, 95, 95, y_low],
                        fill="toself",
                        mode="lines",
                        showlegend=False,
                        line=dict(width=0, color="rgb(150,150,150)"))
    
    fig.add_trace(shape, col=1, row=1)
    
    fig.add_trace(line, col=1, row=1)

    fig.update_yaxes(title="Yield<br>(% of disease free)", row=1, col=1)

    fig.update_layout(standard_layout(True))
    fig.update_layout(annotations=annotz)
    fig.update_layout(legend=dict(x=0.1,
                            y=0.35,
                            bgcolor="rgba(255,255,255,0.5)"))

    fig.show()
    filename = conf_str.replace("/single/", "/single/yield_rf/")
    fig.write_image(filename)



def plot_frequencies(data, conf_str):
    names = list(data['end_of_season'].keys())
    names.remove("SS")
    traces = []

    year = 4

    legend_entries = [f"Start of season {year}",
                       f"End of season {year} (before sexual reproduction)", 
                       f"Start of season {year+1} (after SR step)"]
    
    for key, yr, legend_entry in zip(['start_of_season', 'end_of_season', 'start_of_season'],
                    [year, year, year+1],
                    legend_entries):
        
        y = []
        for ff in names:
            y.append(log10(data[key][ff][int(yr)]))

        bar = go.Bar(x=names,
                        y=y,
                        name=legend_entry
                    )
        
        traces.append(bar)

    fig = go.Figure(data=traces, layout=standard_layout(True))
    fig.update_layout(legend=dict(x=0.4,
                            y=0.1,
                            bgcolor="rgba(255,255,255,0.5)"))
    fig.update_layout(barmode='group')
    fig.update_xaxes(title="Pathogen strain")
    fig.update_yaxes(title="Frequency (log base 10)")
    
    fig.show()
    filename = conf_str.replace("/single/", "/single/strain_freqs/")
    fig.write_image(filename)



def plot_frequencies_over_time(data, conf_str):
    """
    Logit scale, all strain freqs vs year.

    Within season and between season points.
    """
    
    rf_s = data['start_of_season']
    rf_e = data['end_of_season']


    traces = get_strain_freq_traces(rf_s, rf_e)

    fig = go.Figure(data=traces, layout=standard_layout(True))
    fig.update_xaxes(title="Year")
    fig.update_yaxes(title="Frequency (logit scale)")
    
    fig.update_layout(legend=dict(x=0.07,
                        y=1,
                        orientation="h",
                        font=dict(size=18),
                        bgcolor="rgba(255,255,255,0.5)"))

    fig.show()
    filename = conf_str.replace("/single/", "/single/strain_freqs/overtime")
    fig.write_image(filename)



# End of single Tactic

#----------------------------------------------------------------------------------------------
# * Changing dose

def SR_by_dose_plot(data, conf_str):
    rows_list = []

    for key in data.keys():
        my_dict = dict(
            SR = data[key]['selection_vec_dict']['f1'][0],
            RF = float(key.split("rf=")[1]),
            dose = float(key.split("dose=")[1].split(",")[0]),
            yield_ = data[key]['yield_vec'][0],
            )
        rows_list.append(my_dict)
    
    df = pd.DataFrame(rows_list)
    

    traces = []

    all_RFs = list(df.RF.unique())
    all_RFs.reverse()

    for ind, rr in enumerate(all_RFs):
        filt_df = df[(df['RF']==rr) & (df['yield_']>95)]

        min_color = 0
        max_color = 255

        clr = min_color + (max_color - min_color)* ind/(len(all_RFs)-1)

        color = f"rgb({255-clr},{0},{clr})"

        x = np.asarray(filt_df.dose)
        y = np.asarray(filt_df.SR)
        
        line = go.Scatter(x=x, 
                y=y, 
                name=f"Resistance frequency: {str(rr)}",
                line=dict(color=color))

        traces.append(line)
    
    

    text = get_text_annotation(1.05, 1.1, 'Increasing<br>resistance<br>frequency')
    arrow = get_arrow_annotation(1.05, 0.78, 0, 150)
    
    annotz = [text, arrow]

    fig = go.Figure(data=traces, layout=standard_layout(True))
                    
    fig.update_layout(legend=dict(x=0.02, y=1.15, 
                        bgcolor='rgba(255,255,255,0.5)',
                        font=dict(size=14)),
                    annotations=annotz)

    fig.update_xaxes(title="Dose")
    fig.update_yaxes(title="Selection ratio")

    fig.show()
    filename = conf_str.replace("/single/", "/changing_dose/equal_ratio/")
    fig.write_image(filename)

# End of Changing dose

#----------------------------------------------------------------------------------------------
# * Changing fcide

def fcide_grid(x, y, z, filename, labels):
    traces = []

    trace = go.Heatmap(
        x = x,
        y = y,
        z = z,
        colorbar=my_colorbar(labels['cbar'])
    )

    traces.append(trace)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH - 50)

    fig.update_xaxes(title=labels['x'])
    fig.update_yaxes(title=labels['y'])

    fig.show()
    fig.write_image(filename)

# End of Changing fcide

#----------------------------------------------------------------------------------------------
# * Grid of tactics

def dose_grid_heatmap(data, Config, to_plot, conf_str):
    traces = []
    
    x = np.linspace(0, 1, Config.n_doses)
    y = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data[to_plot])

    trace = go.Heatmap(
        x = x,
        y = y,
        z = z,
        colorbar=my_colorbar(TITLE_MAP[to_plot])
    )

    traces.append(trace)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH - 50)

    fig.update_xaxes(title="Dose (fungicide A)")
    fig.update_yaxes(title="Dose (fungicide B)")

    fig.show()
    filename = conf_str.replace("/grid/", f"/grid/dose_grid/{to_plot}")
    fig.write_image(filename)



def dose_grid_RA_heatmap(data, Config, conf_str, yr):
    traces = []
    
    x = np.linspace(0, 1, Config.n_doses)
    y = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data["res_arrays"]["f2"][:,:,yr])

    trace = go.Contour(
        x = x,
        y = y,
        z = z,
        colorbar=my_colorbar("Res. f. B"),
        # colorscale= [(0,"yellow"), (1,"red")],
        # range_color=[0,1]
    )

    traces.append(trace)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH - 50)

    fig.update_xaxes(title="Dose (fungicide A)")
    fig.update_yaxes(title="Dose (fungicide B)")

    fig.show()
    filename = conf_str.replace("/grid/", f"/grid/dose_grid/res_array_{yr}")
    fig.write_image(filename)





def dose_sum_hobb_vs_me(data, Config, to_plot, conf_str):
    """
    Lines on heatmap, and then log ratio of RFs at break down
    """

    fig = make_subplots(rows=2, cols=2, horizontal_spacing=0.2)

    heatmap_subplot = []
    
    xheat = np.linspace(0, 1, Config.n_doses)
    yheat = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data[to_plot])
    
    # for legend
    for name_, clr, dash_ in zip(["Equal resistance<br>at breakdown", 
                                        "First year selection"],
                                ["blue", "black"],
                                ["dash", "dot"]):
        heatmap_subplot.append(go.Scatter(
            x=[-0.3,-0.31],
            y=[-0.3,-0.31],
            line=dict(color=clr, dash=dash_),
            mode="lines",
            name=name_
        ))
    
    clrbar = my_colorbar(TITLE_MAP[to_plot])
    clrbar.update(dict(x=0.42, y=0.79, len=0.43))

    heatmap = go.Heatmap(
        x = xheat,
        y = yheat,
        z = z,
        colorscale=grey_colorscale(z),
        colorbar=clrbar
    )

    heatmap_subplot.append(heatmap)
    

    # add lines on heatmap
    N_y_int = 2*Config.n_doses-1
    y_intrcpt = np.linspace(0,2,N_y_int)
    trc_out, colors, inds_list = get_heatmap_lines(Config, z, y_intrcpt)

    heatmap_subplot += trc_out

    my_strat_traces, RBF_diff = get_RFB_diff_traces(data, z, N_y_int, inds_list, colors)
    hobb_strat_traces, eq_fy = get_eq_sel_traces(data, z, N_y_int, inds_list, colors)
    
    for trace in my_strat_traces:
        fig.add_trace(trace, row=1, col=2)
    
    for trace in hobb_strat_traces:
        fig.add_trace(trace, row=2, col=2)
    
    eq_contour = contour_at_0(xheat, yheat, RBF_diff, "blue", "dash")
    eq_fy_contour = contour_at_0(xheat, yheat, eq_fy, "black", "dot")


    heatmap_subplot.append(eq_contour)
    heatmap_subplot.append(eq_fy_contour)

    # col 1, row 1
    for trace in heatmap_subplot:
        fig.add_trace(trace, row=1, col=1)
    
    

    annotz = []

    annot1 = get_text_annotation(0.58, -0.025, 'More selection<br>for f. B')
    annot2 = get_text_annotation(1.02, -0.025, 'More selection<br>for f. A')
    arrow1 = get_arrow_annotation(0.96, -0.04, -200, 0)
    arrow2 = get_arrow_annotation(0.64, -0.04,  200, 0)
    
    annot1b = get_text_annotation(0.58,  0.55, 'More resistance<br>to f. B')
    annot2b = get_text_annotation(1.02,  0.55, 'More resistance<br>to f. A')
    arrow1b = get_arrow_annotation(0.96, 0.535, -200, 0)
    arrow2b = get_arrow_annotation(0.64, 0.535,  200, 0)
    
    annotz += [annot1, annot2, arrow1, arrow2, annot1b, annot2b, arrow1b, arrow2b]


    fig.update_layout(standard_layout(True))
    fig.update_layout(width=2*PLOT_WIDTH, 
                        height=2*PLOT_WIDTH - 50,
                        annotations=annotz,
                        legend=dict(
                                    x=0.25,
                                    y=0.25,
                                    yanchor="middle",
                                    xanchor="center",
                                    font=dict(size=18)
                                    ),
                        font=dict(size=18)
                                    )

    fig.update_xaxes(title="Difference in logit of<br>res. freqs. at breakdown", row=1, col=2, showgrid=False, zeroline=False)
    fig.update_xaxes(title="Difference of log of<br>selection ratios after one year", row=2, col=2, showgrid=False, zeroline=False)
    fig.update_yaxes(title="Effective life", row=1, col=2)
    fig.update_yaxes(title="Effective life", row=2, col=2)
    
    # if heatmap not contour use [0-dx, 1+dx] etc
    # is order correct? shape[0]/[1]
    # dx = 0.5*(1/(-1+z.shape[1]))
    # dy = 0.5*(1/(-1+z.shape[0]))
    dx = 0.01
    dy = 0.01
    
    fig.update_xaxes(title="Dose (fungicide A)", range=[0-dx,1+dx], row=1, col=1, showgrid=False, zeroline=False)
    fig.update_yaxes(title="Dose (fungicide B)", range=[0-dy,1+dy], row=1, col=1, showgrid=False, zeroline=False)

    fig.show()
    filename = conf_str.replace("/grid/", "/dose_space/dose_sum_hobb_vs_me/")
    fig.write_image(filename)








def dose_sum_LR(data, Config, to_plot, conf_str):
    """
    Lines on heatmap, and then log ratio of RFs at break down
    """

    fig = make_subplots(rows=1, cols=2, horizontal_spacing=0.2)

    my_strat_traces = []
    heatmap_subplot = []
    
    xheat = np.linspace(0, 1, Config.n_doses)
    yheat = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data[to_plot])
    
    heatmap = go.Heatmap(
        x = xheat,
        y = yheat,
        z = z,
        colorscale=grey_colorscale(z),
        colorbar=my_colorbar(TITLE_MAP[to_plot])
    )

    heatmap_subplot.append(heatmap)

    for name_, clr, dash_ in zip(["Equal resistance<br>at breakdown"],
                                ["blue"],
                                ["dash"]):
        heatmap_subplot.append(go.Scatter(
            x=[-0.3,-0.31],
            y=[-0.3,-0.31],
            line=dict(color=clr, dash=dash_),
            mode="lines",
            name=name_
        ))


    

    # add lines on heatmap
    N_y_int = 2*Config.n_doses-1
    y_intrcpt = np.linspace(0,2,N_y_int)
    trc_out, colors, inds_list = get_heatmap_lines(Config, z, y_intrcpt)

    heatmap_subplot += trc_out

    
    my_strat_traces, RBF_diff = get_RFB_diff_traces(data, z, N_y_int, inds_list, colors)
    
    
    for trace in my_strat_traces:
        fig.add_trace(trace, row=1, col=1)
    
    eq_contour = contour_at_0(xheat, yheat, RBF_diff, "blue", "dash")

    heatmap_subplot.append(eq_contour)

    for trace in heatmap_subplot:
        fig.add_trace(trace, row=1, col=2)
    
    annotz = []

    # for x_pos, y_pos, text, show_arr, arrow_length in zip(
    #         [0, 0.4, 0.34, 0.06],
    #         [-0.06, -0.06, -0.08, -0.08],
    #         ['More resistance<br>to f. B', 'More resistance<br>to f. A', '', ''],
    #         [False, False, True, True],
    #         [None, None, -200, 200]):
    #     annotz.append(dict(
    #         x=x_pos,
    #         y=y_pos,
    #         text=text,

    #         showarrow=show_arr,
    #         arrowcolor=LABEL_COLOR,
    #         arrowsize=2,
    #         arrowwidth=1,
    #         arrowhead=2,
            
    #         ax=arrow_length,
    #         ay=0,
                
    #         xref='paper',
    #         yref='paper',

    #         xanchor="center",
    #         yanchor="top",

    #         font=dict(
    #                 size=14,
    #                 color=LABEL_COLOR,
    #             ),
    #     ))

    annot1 = get_text_annotation(0, -0.06, 'More resistance<br>to f. B')
    annot2 = get_text_annotation(0.4, -0.06, 'More resistance<br>to f. A')
    arrow1 = get_arrow_annotation(0.34, -0.08, -200, 0)
    arrow2 = get_arrow_annotation(0.06, -0.08,  200, 0)
    
    annotz += [annot1, annot2, arrow1, arrow2]


    fig.update_layout(standard_layout(True))
    fig.update_layout(width=2*PLOT_WIDTH, 
                        height=PLOT_WIDTH - 50,
                        annotations=annotz,
                        legend=dict(
                                    x=0.36,
                                    y=0.95,
                                    yanchor="top",
                                    font=dict(size=14)
                                    ),
                        font=dict(size=20)
                                    )

    fig.update_xaxes(title="Difference in logit of<br>resistance frequencies<br>at breakdown", row=1, col=1, showgrid=False, zeroline=False)
    fig.update_yaxes(title="Effective life", row=1, col=1)
    
    # if heatmap not contour use [0-dx, 1+dx] etc
    # is order correct? shape[0]/[1]
    # dx = 0.5*(1/(-1+z.shape[1]))
    # dy = 0.5*(1/(-1+z.shape[0]))
    dx = 0.01
    dy = 0.01
    
    fig.update_xaxes(title="Dose (fungicide A)", range=[0-dx,1+dx], row=1, col=2, showgrid=False, zeroline=False)
    fig.update_yaxes(title="Dose (fungicide B)", range=[0-dy,1+dy], row=1, col=2, showgrid=False, zeroline=False)

    fig.show()
    filename = conf_str.replace("/grid/", f"/dose_space/dose_sum/{to_plot}")
    fig.write_image(filename)


# End of grid of tactics


#----------------------------------------------------------------------------------------------
# * Dose space

def dose_space_contour(data, to_plot, conf_str):
    traces = []
    
    x = [log2(xx) for xx in data['contours_radial']]
    # y = [log2(xx) for xx in data['contour_perp']]
    y = data['contour_perp']

    z = data[to_plot]

    trace = go.Heatmap(
        x = x,
        y = y,
        z = z,
        colorbar=my_colorbar(TITLE_MAP[to_plot])
    )

    traces.append(trace)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH-50)

    spaces = max(floor(len(x)/5),1)
    xticks = x[0:len(x):spaces]
    xtick_text = list(map(str, data['contours_radial']))[0:len(x):spaces]

    fig.update_xaxes(title="Ratio (log scale)",
        tickvals=xticks,
        ticktext=xtick_text)


    fig.update_yaxes(title="'Strength'",
        showgrid=False,
        )

    fig.show()
    filename = conf_str.replace("/grid/", "/dose_space/yield_by_year/")
    fig.write_image(filename)


def radial(radial_data, grid_data, Config):
    
    fig = make_subplots(rows=1, cols=2, horizontal_spacing=0.2)

    x = np.linspace(0, 1, Config.n_doses)
    y = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(grid_data['FY'])

    heatmap = go.Contour(
        x = x,
        y = y,
        z = z,
        colorscale=grey_colorscale(z),
        colorbar=dict(
            title = TITLE_MAP['FY'],
            titleside = 'right',
        )
        )

    fig.add_trace(heatmap, row=1, col=2)

    angles = list(radial_data.angle.unique())
    
    for ind, angle in enumerate(angles):
        data_use = radial_data[(radial_data["angle"]==angle) & (radial_data['FY']>0)]
        
        xx = list(data_use.d1)
        yy = list(data_use.d2)
        
        value = 255*ind/(len(angles)-1)

        clr = f"rgb(0,{value},{255-value})"

        heatmap_line = go.Scatter(x=xx,
                            y=yy,
                            line=dict(color=clr),
                            showlegend=False
                            )
        fig.add_trace(heatmap_line, row=1, col=2)


        x = list(data_use.radius)
        y = list(data_use.FY)
        
        ang = 2*90*angle/pi
        ang = round(ang, 0)

        name_string = f"Angle: {ang}" + u"\u00B0"
        
        line = go.Scatter(x=x,
                        y=y,
                        line=dict(color=clr),
                        name= name_string
                        )

        fig.add_trace(line, row=1, col=1)

    
    annotz = []

    # for x_pos, y_pos, text, show_arr, arrow_length in zip(
    #         [0.4, 0.34],
    #         [-0.055, -0.08],
    #         ['Stronger fungicide<br>mixture', ''],
    #         [False, True],
    #         [None, -200]):
    #     annotz.append(dict(
    #         x=x_pos,
    #         y=y_pos,
    #         text=text,

    #         showarrow=show_arr,
    #         arrowcolor=LABEL_COLOR,
    #         arrowsize=2,
    #         arrowwidth=1,
    #         arrowhead=2,
            
    #         ax=arrow_length,
    #         ay=0,
                
    #         xref='paper',
    #         yref='paper',

    #         xanchor="center",
    #         yanchor="top",

    #         font=dict(
    #                 size=14,
    #                 color=LABEL_COLOR,
    #             ),
    #     ))
    
    annot1 = get_text_annotation(0.4, -0.055, 'Stronger fungicide<br>mixture')
    arrow1 = get_arrow_annotation(0.34, -0.08, -200, 0)
    
    annotz += [annot1, arrow1]


    fig.update_layout(standard_layout(True))
    fig.update_layout(width=2*PLOT_WIDTH,
                        height=2*PLOT_WIDTH - 50,
                        legend=dict(x=0.35,
                                    y=0.2,
                                    yanchor="middle",
                                    font=dict(size=14)
                                    ),
                        font=dict(size=20),
                        annotations=annotz
                        )


    fig.update_xaxes(title="Mixture strength<br>(radius measured from origin)", row=1, col=1)
    fig.update_yaxes(title="Effective life", row=1, col=1)

    dx = 0.01
    dy = 0.01

    fig.update_xaxes(title="Dose (fungicide A)", range=[0-dx,1+dx], row=1, col=2, showgrid=False)
    fig.update_yaxes(title="Dose (fungicide B)", range=[0-dy,1+dy], row=1, col=2, showgrid=False)

    fig.show()
    conf_str = Config.config_string_img
    filename = conf_str.replace("/grid/", "/dose_space/radial/")
    fig.write_image(filename)




def first_year_yield(data, Config):
    
    traces = []
    
    x = np.linspace(0, 1, Config.n_doses)
    y = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data['yield_array'][:,:,0])

    trace = go.Contour(
        x = x,
        y = y,
        z = z,
        colorbar=dict(
            title = 'Yield',
            titleside = 'right',
        )
    )

    traces.append(trace)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH-50)

    fig.update_xaxes(title="Dose (fungicide A)",
        showgrid=False)

    fig.update_yaxes(title="Dose (fungicide B)",
        showgrid=False,
        )

    fig.show()
    conf_str = Config.config_string_img
    filename = conf_str.replace("/grid/", "/dose_space/first_year_yield/")
    fig.write_image(filename)


def eq_RFB_contours(data, Config, title=None):
    
    traces = get_multi_contour_traces(data, Config)

    fig = go.Figure(data=traces, layout=standard_layout(False))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH-50)
    # dy = 0.5*(1/(-1+z.shape[0]))
    dx = 0.01
    dy = 0.01

    fig.update_xaxes(title="Dose (fungicide A)",
        range=[0-dx,1+dx],
        showgrid=False)

    fig.update_yaxes(title="Dose (fungicide B)",
        range=[0-dy,1+dy],
        showgrid=False)

    fig.update_layout(title=title)

    fig.show()
    conf_str = Config.config_string_img
    filename = conf_str.replace("/grid/", "/dose_space/eq_RFB_contours/")
    fig.write_image(filename)



def MS_RFB_scatter_plot(data, Config):

    traces = get_MS_RFB_traces(data)

    fig = go.Figure(data=traces, layout=standard_layout(True))

    fig.update_layout(width=PLOT_WIDTH, height=PLOT_WIDTH-50)

    fig.update_layout(legend=dict(x=1.05,
                            y=1.1,
                            yanchor="top",
                            xanchor="right",
                            font=dict(size=16),
                            # bgcolor="rgba(255,255,255,0.5)"
                            ))
    

    fig.update_xaxes(title=u"\u0394" + "<i><sub>RFB</sub></i>")
    fig.update_yaxes(title="Dose sum", range=[0.3, 2.04])

    fig.show()
    conf_str = Config.config_string_img
    filename = conf_str.replace("/grid/", "/dose_space/MS_RFB_scatter_plot/")
    fig.write_image(filename)




# End of Dose space


# * RF Ratio


def dose_grid_heatmap_with_contours(data, Config, contours, conf_str):
    traces = []
    
    x = np.linspace(0, 1, Config.n_doses)
    y = np.linspace(0, 1, Config.n_doses)

    z = np.transpose(data["FY"])

    trace = go.Heatmap(
        x = x,
        y = y,
        z = z,
        colorbar=my_colorbar(TITLE_MAP["FY"])
    )

    traces.append(trace)
    
    names = ["Equal R.Fs<br>at breakdown", "Equal selection<br>in first year"]
    for cont_df, name in zip(contours, names):
        xx = cont_df.f1
        yy = cont_df.f2

        scatter = go.Scatter(x=xx, y=yy, name=name)

        traces.append(scatter)

    fig = go.Figure(data=traces, layout=standard_layout(True))

    fig.update_layout(width=PLOT_WIDTH, 
                    height=PLOT_WIDTH-50,
                    legend=dict(x=1.25,
                                y=1,
                                yanchor="top",
                                xanchor="left",
                                font=dict(size=16)
                                ))

    fig.update_xaxes(title="Dose (fungicide A)")
    fig.update_yaxes(title="Dose (fungicide B)")

    fig.show()
    filename = conf_str.replace("/grid/", "/rf_ratio/dose_grid_and_contours/")
    fig.write_image(filename)


def outcomes_by_ratio(data, conf_str):
    traces = []
    
    x = [log10(d) for d in data.ratio]
    yEqS = data.EqS
    yRFB = data.RFB

    y_list = [yRFB, yEqS]
    names = ["STRATEGY: Equal resistance frequencies at breakdown",
            "STRATEGY: Equal selection in first year"
            ]

    for yy, name in zip(y_list, names):
        line = go.Scatter(
            x = x,
            y = yy,
            name=name
            )

        traces.append(line)
    
    
    fig = go.Figure(data=traces, layout=standard_layout(True))

    fig.update_layout(width=2*PLOT_WIDTH,
                    height=PLOT_WIDTH,
                    legend=dict(x=0.3,
                                y=1,
                                yanchor="top",
                                xanchor="left",
                                font=dict(size=16)
                                ))
    xticks = list(x[::2])
    
    if 0 not in xticks:
        xticks.append(0)


    xtick_text = [10**(x) for x in xticks]

    fig.update_xaxes(title="Ratio of initial resistance frequencies",
            tickvals=xticks,
            ticktext=xtick_text
            )
    fig.update_yaxes(title="Effective lifetime")

    fig.show()
    
    x_str = [str(int(i)) for i in x]
    ratio_string = "".join(x_str)
    filename = conf_str.replace("/grid/", f"/rf_ratio/outcomes/{ratio_string}")
    fig.write_image(filename)

# End of RF Ratio



# Paper Figs
    
class DiseaseProgressCurvesAll:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 650

        self.xx = data.states_list[0].t
        
        self.states_list = data.states_list

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)




    def _generate_figure(self):

        traces_dict = self.get_model_output_overview_traces()

        ugly_fig = self.add_traces_to_layout_model_output_overview(traces_dict)

        fig = self._sort_layout(ugly_fig)

        return fig
        
    




    def get_model_output_overview_traces(self):
        S_R_traces = self.get_S_R_traces()
        E_traces = self.get_E_traces()
        I_traces = self.get_I_traces()
        F_traces = self.get_F_traces()

        out = dict(S_R = S_R_traces,
                    E = E_traces,
                    I = I_traces,
                    F = F_traces)

        return out
    



    def get_S_R_traces(self):
        
        out = []

        for key in ['S', 'R']:
            out.append(self.get_DPC_trace(key))
        return out
    
    

    def get_E_traces(self):
        
        out = []

        for key in ['ERR', 'ERS', 'ESR', 'ESS']:
            out.append(self.get_DPC_trace(key))

        return out
    


    
    def get_I_traces(self):
        
        out = []

        for key in ['IRR', 'IRS', 'ISR', 'ISS']:
            out.append(self.get_DPC_trace(key))

        return out
    


    def get_F_traces(self):

        out = []

        for key in ['F1', 'F2']:
            out.append(self.get_DPC_trace(key))

        return out



    def get_DPC_trace(self, key):
        clr = ATTRS_DICT[key]['color']
        dash = ATTRS_DICT[key]['dash']
        name = ATTRS_DICT[key]['name']
        # ind = ATTRS_DICT[key]['ind']

        yy = vars(self.states_list[0])[key]

        return go.Scatter(x=self.xx,
                    y=yy,
                    line=dict(color=clr, dash=dash),
                    name=name
                    )






    def add_traces_to_layout_model_output_overview(self, data_dict):
        fig = make_subplots(rows=2, cols=2, horizontal_spacing=0.2)

        self.add_traces(fig, data_dict['S_R'], 1, 1)
        self.add_traces(fig, data_dict['E'], 1, 2)
        self.add_traces(fig, data_dict['I'], 2, 1)
        self.add_traces(fig, data_dict['F'], 2, 2)

        return fig



    def _sort_layout(self, fig):
        fig = self.update_axes(fig)


        fig.update_layout(standard_layout(False, self.width, self.height))
        
        fig = self.add_corner_text_labels(fig)

        fig = self.sort_legend(fig)

        return fig


    def sort_legend(self, fig):
        fig.update_layout(showlegend=True, legend=dict(font=dict(size=16)))
        return fig



    def add_corner_text_labels(self, fig):
        top_row = 1.08
        bottom_row = 0.5
        
        left = 0.01
        middle = 0.58

        c1 = get_big_text_annotation(left, top_row, 'A')
        c2 = get_big_text_annotation(middle, top_row, 'B')
        c3 = get_big_text_annotation(left, bottom_row, 'C')
        c4 = get_big_text_annotation(middle, bottom_row, 'D')
        
        annotz = [c1, c2, c3, c4]

        fig.update_layout(annotations=annotz)
        return fig




    def add_traces(self, fig, traces, row, col):
        for trace in traces:
            fig.add_trace(trace, row, col)

        return fig




    def update_axes(self, fig):
        fig.update_xaxes(row=1, col=1, showgrid=False)
        fig.update_xaxes(row=1, col=2, showgrid=False)
        
        fig.update_xaxes(title="Time (degree-days)", row=2, col=1, showgrid=False)
        fig.update_xaxes(title="Time (degree-days)", row=2, col=2, showgrid=False)
        
        fig.update_yaxes(title="L.A.I.", row=1, col=1)
        fig.update_yaxes(title="L.A.I.", row=1, col=2)
        fig.update_yaxes(title="L.A.I.", row=2, col=1)
        fig.update_yaxes(title="Concentration", row=2, col=2)

        return fig
    


    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/single/", "/paper_figs/model_overview_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)










class CombinedModelPlot:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 950
        
        self.states_list = data.states_list

        self.data = data

        self.DPC_year = 5

        self.xx = data.states_list[self.DPC_year].t

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)




    def _generate_figure(self):

        m_o_trcs_dict = self.get_model_output_overview_traces()

        y_rf_trcs_dict = self.get_yield_RF_traces()

        traces_dict = {**m_o_trcs_dict, **y_rf_trcs_dict}

        ugly_fig = self.add_traces_to_layout(traces_dict)

        fig = self._sort_layout(ugly_fig)

        return fig
        
    




    def get_model_output_overview_traces(self):
        S_R_traces = self.get_S_R_traces()
        E_traces = self.get_E_traces()
        I_traces = self.get_I_traces()
        F_traces = self.get_F_traces()

        out = dict(S_R = S_R_traces,
                    E = E_traces,
                    I = I_traces,
                    F = F_traces)

        return out
    




    def get_S_R_traces(self):
        
        out = []

        for key in ['S', 'R']:
            out.append(self.get_DPC_trace(key))
        return out
    
    

    def get_E_traces(self):
        
        out = []

        for key in ['ERR', 'ERS', 'ESR', 'ESS']:
            out.append(self.get_DPC_trace(key))

        return out
    


    
    def get_I_traces(self):
        
        out = []

        for key in ['IRR', 'IRS', 'ISR', 'ISS']:
            out.append(self.get_DPC_trace(key))

        return out
    


    def get_F_traces(self):

        out = []

        for key in ['fung_1', 'fung_2']:
            out.append(self.get_DPC_trace(key))

        return out



    def get_DPC_trace(self, key):
        clr = ATTRS_DICT[key]['color']
        dash = ATTRS_DICT[key]['dash']
        name = ATTRS_DICT[key]['name']

        return go.Scatter(x=self.xx,
                    y=vars(self.states_list[self.DPC_year])[key],
                    line=dict(color=clr, dash=dash),
                    legendgroup="DPC",
                    name=name
                    )


    def get_yield_RF_traces(self):        

        yield_traces = self.get_yield_traces()
        
        RF_traces = self.get_RF_traces()
        
        traces = {"yield": yield_traces, "RF": RF_traces}

        return traces
    
    
    def get_yield_traces(self):
        out = []

        yy = self.data.yield_vec
        
        xx = list(range(1,1+len(yy)))

        line = go.Scatter(x=xx, y=yy, name="Yield", legendgroup="YRF")
        
        Y_LOW = yy[-1]-2

        self.yield_lower_lim = Y_LOW
        
        X_END = 0.5 + xx[-1]
                
        shape = go.Scatter(x=[0, 0, X_END, X_END],
                            y=[Y_LOW, 95, 95, Y_LOW],
                            fill="toself",
                            mode="lines",
                            showlegend=False,
                            line=dict(width=0, color="rgb(150,150,150)"))
        
        out.append(shape)
        out.append(line)

        return out
    




    def get_RF_traces(self):
        out = []
        
        y1 = self.data.res_vec_dict['f1']
        y2 = self.data.res_vec_dict['f2']
        
        xx = list(range(len(y1)))

        for data, name, dash, col in zip([y1, y2], ['A', 'B'], ['solid', 'dot'], ['red', 'blue']):
            line = go.Scatter(x=xx, 
                y=data,
                name=f"R.F. (fung. {name})",
                legendgroup="YRF",
                line=dict(dash=dash, color=col))
            out.append(line)

        return out



    def add_traces_to_layout(self, data_dict):
        fig = make_subplots(rows=2, cols=3, 
                            horizontal_spacing=0.2,
                            shared_xaxes=True,
                            # row_heights=[0.3, 0.3]
                            )

        self.add_traces(fig, data_dict['S_R'], 1, 1)
        self.add_traces(fig, data_dict['E'], 1, 2)
        self.add_traces(fig, data_dict['I'], 2, 1)
        self.add_traces(fig, data_dict['F'], 2, 2)
        
        self.add_traces(fig, data_dict['yield'], 1, 3)
        self.add_traces(fig, data_dict['RF'], 2, 3)

        return fig



    def _sort_layout(self, fig):
        fig = self.update_axes(fig)

        fig.update_layout(standard_layout(False, self.width, self.height))
        
        annotz = self.get_corner_text_labels_comb()
        annotz += self.get_unacceptable_yield_annotation()
        
        fig.update_layout(annotations=annotz)

        fig = self.add_diagram(fig)

        fig.update_layout(margin=dict(t=450))

        fig = self.sort_legend(fig)

        return fig


    def add_diagram(self, fig):
        
        img = Image.open("create_figs/img/diagram.png")

        fig.add_layout_image(
            dict(
                source=img,
                
                xref="paper", yref="paper",
                x=0,
                y=1.15,
                sizex=1.5,
                sizey=1.5,
                xanchor="left",
                yanchor="bottom",
                opacity=1,
                ))
        return fig


    def sort_legend(self, fig):
        fig.update_layout(showlegend=True, legend=dict(font=dict(size=16)))
        return fig


    @staticmethod
    def get_corner_text_labels_comb():

        very_top = 1.8
        
        top_row = 1.12
        bottom_row = 0.54
        
        left = -0.04
        middle = 0.33
        right = 0.72

        annotz = [
            get_big_text_annotation(left, very_top, 'A'),
            get_big_text_annotation(left, top_row, 'B'),
            get_big_text_annotation(left, bottom_row, 'C'),
            get_big_text_annotation(middle, top_row, 'D'),
            get_big_text_annotation(middle, bottom_row, 'E'),
            get_big_text_annotation(right, top_row, 'F'),
            get_big_text_annotation(right, bottom_row, 'G'),
            ]

        return annotz



    def get_unacceptable_yield_annotation(self):
        # 0.5*(95+self.yield_lower_lim)

        return [dict(
            xref="x3",
            # yref="y1",
            xanchor="left",
            # yanchor="top",
            x= 0.5,
            y= 0.55,
            text="Unacc.<br>yield",
            showarrow=False,
            )]



    def add_traces(self, fig, traces, row, col):
        for trace in traces:
            fig.add_trace(trace, row, col)

        return fig




    def update_axes(self, fig):

        fig.update_xaxes(row=1, col=1, showgrid=False)
        fig.update_xaxes(row=1, col=2, showgrid=False)
        
        fig.update_xaxes(title="Time<br>(degree-days)", row=2, col=1, showgrid=False, zeroline=False)
        fig.update_xaxes(title="Time<br>(degree-days)", row=2, col=2, showgrid=False, zeroline=False)
        
        fig.update_xaxes(title="Time (years)", row=2, col=3, showgrid=False, zeroline=False)
        fig.update_xaxes(row=1, col=3, showgrid=False, zeroline=False)
        
        
        fig.update_yaxes(title="L.A.I.", row=1, col=1, showgrid=False, zeroline=False)
        fig.update_yaxes(title="L.A.I.", row=1, col=2, showgrid=False, zeroline=False)
        fig.update_yaxes(title="L.A.I.", row=2, col=1, showgrid=False, zeroline=False)
        fig.update_yaxes(title="Concentration", row=2, col=2, showgrid=False, zeroline=False)

        fig.update_yaxes(title="Yield", row=1, col=3, showgrid=False, zeroline=False)
        fig.update_yaxes(title="R.F.", row=2, col=3, showgrid=False, zeroline=False)

        return fig
    


    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/single/", "/paper_figs/model_overview_combined_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)


# End of CombinedModelPlot














class DoseSpaceScenariosPlot:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 650

        self.data = data

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)



    def _generate_figure(self):
        traces = self._get_traces()

        ugly_fig = self._add_traces_to_figure(traces)

        fig = self._sort_layout(ugly_fig)

        return fig



    def _get_traces(self):        
        traces = []

        traces.append(self.get_ERFB_legend_entry())
        traces.append(self.get_EqSel_legend_entry())

        traces.append(self.get_FY_heatmap())
        
        traces.append(self.get_ERFB_contour())
        traces.append(self.get_EqSel_contour())
        
        traces.append(self.get_full_dose_point())
        traces.append(self.get_min_dose_point())
        

        return traces


    def get_EqSel_legend_entry(self):
        return go.Scatter(x=[1], 
                    y=[1],
                    mode="lines",
                    line=dict(color="blue", dash="dot"),
                    name="ES contour"
                    )


    def get_ERFB_legend_entry(self):
        return go.Scatter(x=[1], 
                    y=[1],
                    mode="lines",
                    line=dict(color="black", dash="solid"),
                    name="ERFB contour"
                    )



    def get_FY_heatmap(self):
        FYs = np.transpose(self.data.FY)

        xheat = np.linspace(0, 1, FYs.shape[0])
        yheat = np.linspace(0, 1, FYs.shape[1])

        heatmap = go.Heatmap(
            x = xheat,
            y = yheat,
            z = FYs,
            colorscale = grey_colorscale(FYs),
            colorbar = my_colorbar(TITLE_MAP["FY"])
            )

        return heatmap


    def get_ERFB_contour(self):
        z = EqualResFreqBreakdownArray(self.data).array

        x = np.linspace(0, 1, z.shape[0])
        y = np.linspace(0, 1, z.shape[1])

        z_transpose = np.transpose(z)

        out = contour_at_0(x, y, z_transpose, 'black', 'solid')
        out['name'] = "Delta RFB"

        return out



    def get_EqSel_contour(self):
        z = EqualSelectionArray(self.data).array

        x = np.linspace(0, 1, z.shape[0])
        y = np.linspace(0, 1, z.shape[1])

        z_transpose = np.transpose(z)

        out = contour_at_single_level(x, y, z_transpose, 0.5, 'blue', 'dot')
        out['name'] = "Equal Selection"

        return out



    @staticmethod
    def get_full_dose_point():
        out = go.Scatter(x=[1],
                y=[1],
                marker=dict(color='red', size=16),
                marker_symbol='circle',
                mode='markers',
                name="Full dose",
                )

        return out



    def get_min_dose_point(self):
        FYs = np.transpose(self.data.FY)

        x = np.linspace(0, 1, FYs.shape[0])
        
        minEqDoseELVec = np.asarray([float(FYs[i, i]) for i in range(FYs.shape[0])])

        ind = np.where(minEqDoseELVec>1)

        min_val = x[int(ind[0][0])]

        out = go.Scatter(x=[min_val],
                y=[min_val],
                marker=dict(color='green', size=16),
                mode='markers',
                marker_symbol='square',
                name="Min. dose",
                )

        return out




    def _add_traces_to_figure(self, traces):
        fig = go.Figure(data=traces, layout=standard_layout(True, self.width, self.height))
        return fig



    def _sort_layout(self, fig):
        fig = self._update_axes(fig)
        fig = self._update_legend(fig)
        return fig
    


    @staticmethod
    def _update_axes(fig):
        eps = 0.04
        fig.update_xaxes(title="Dose (fungicide A)", range=[0-eps,1+eps], showgrid=False, zeroline=False)
        fig.update_yaxes(title="Dose (fungicide B)", range=[0-eps,1+eps], showgrid=False, zeroline=False)
        return fig
    



    @staticmethod
    def _update_legend(fig):
        fig.update_layout(legend=dict(x=1.2, y=0.95))
        return fig
    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/grid/", "/paper_figs/dose_space_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)





class DosesScatterPlot:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 620

        self.data = data

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)



    def _generate_figure(self):
        traces = self._get_traces()

        ugly_fig = self._add_traces_to_figure(traces)

        fig = self._sort_layout(ugly_fig)

        return fig



    def _get_traces(self):        
        traces = []
        
        line = go.Scatter(x=[0,0],
                    y=[0,12.5],
                    line=dict(color='rgb(50,50,50)', dash='dot'),
                    mode="lines"
                    )
        
        traces.append(line)

        z = EqualResFreqBreakdownArray(self.data).array
        FYs = self.data.FY

        x = np.asarray(z).flatten()
        y = np.asarray(FYs).flatten()

        dose_sum_cols = self.get_dose_sum_vec(z.shape, y)
        

        scatter = go.Scatter(
                x=x,
                y=y,
                mode="markers",
                text=dose_sum_cols,
                marker=dict(color=dose_sum_cols,
                    size=6,
                    line=dict(width=0.2,
                            color='black'),

                    colorbar=dict(title="Sum of doses"),
                    colorscale='Viridis',
                    showscale=True)
                )

        traces.append(scatter)

        return traces



    def get_dose_sum_vec(self, matrix_shape, FYs_flat):
        array = np.zeros(matrix_shape)

        for i in range(matrix_shape[0]):
            for j in range(matrix_shape[1]):
                array[i,j] = i+j
        
        array = array*(2/array[-1,-1])

        
        ds_cols = array.flatten()

        dose_sum_cols = [ds_cols[i] if FYs_flat[i] else "grey" for i in range(len(ds_cols))]

        return dose_sum_cols



    def _add_traces_to_figure(self, traces):
        fig = go.Figure(data=traces, layout=standard_layout(False, self.width, self.height))
        return fig



    def _sort_layout(self, fig):
        fig = self._update_axes(fig)
        return fig
    


    @staticmethod
    def _update_axes(fig):
        fig.update_xaxes(title=r"$\Delta_{RFB}$")
        fig.update_yaxes(title="Effective life")
        return fig

    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/grid/", "/paper_figs/doses_scatter_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)





class YieldAndRfPlot:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 620

        self.data = data

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)



    def _generate_figure(self):
        trace_dict = self.get_trace_dict()

        ugly_fig = self._add_traces_to_figure(trace_dict)

        fig = self._sort_layout(ugly_fig)

        return fig



    def get_trace_dict(self):        
        

        yield_traces = self.get_yield_traces()
        
        RF_traces = self.get_RF_traces()
        
        traces = {"yield": yield_traces, "RF": RF_traces}

        return traces





    def get_yield_traces(self):
        out = []

        yy = self.data.yield_vec
        
        xx = list(range(1,1+len(yy)))

        line = go.Scatter(x=xx, y=yy, name="Yield")
        
        Y_LOW = yy[-1]-2

        self.yield_lower_lim = Y_LOW
        
        X_END = 0.5 + xx[-1]
                
        shape = go.Scatter(x=[0, 0, X_END, X_END],
                            y=[Y_LOW, 95, 95, Y_LOW],
                            fill="toself",
                            mode="lines",
                            showlegend=False,
                            line=dict(width=0, color="rgb(150,150,150)"))
        
        out.append(shape)
        out.append(line)

        return out
    




    def get_RF_traces(self):
        out = []
        
        y1 = self.data.res_vec_dict['f1']
        y2 = self.data.res_vec_dict['f2']
        
        xx = list(range(len(y1)))

        for data, name, dash in zip([y1, y2], ['A', 'B'], ['dot', 'solid']):
            line = go.Scatter(x=xx, 
                y=data,
                name=f"<br>Resistance<br>frequency<br>(fungicide {name})",
                line=dict(dash=dash))
            out.append(line)

        return out


    def _add_traces_to_figure(self, trace_dict):
        fig = make_subplots(rows=2, cols=1, shared_xaxes=True)
        
        fig = self._add_yield_traces(fig, trace_dict['yield'])

        fig = self._add_RF_traces(fig, trace_dict['RF'])
        
        return fig



    @staticmethod
    def _add_yield_traces(fig, traces):
        for trace in traces:
            fig.add_trace(trace, row=1, col=1)
        return fig
    


    @staticmethod
    def _add_RF_traces(fig, traces):
        for trace in traces:
            fig.add_trace(trace, row=2, col=1)
        return fig






    def _sort_layout(self, fig):
        fig = self._update_layout(fig)
        fig = self._update_axes(fig)
        return fig
    
    
    
    def _update_layout(self, fig):
        fig.update_layout(standard_layout(True, self.width, self.height))
        
        text = self.get_unacceptable_yield_annotation()
        corners = self._get_corner_text_labels()
        annotz = corners + text
        
        fig.update_layout(annotations=annotz)
        return fig


    def get_unacceptable_yield_annotation(self):
        return [dict(
            xref="x1",
            yref="y1",
            xanchor="left",
            x= 1,
            y= 2 + 0.5*(95+self.yield_lower_lim),
            text="Unacceptable yield",
            showarrow=False,
            )]



    @staticmethod
    def _update_axes(fig):
        fig.update_xaxes(title="Time (years)", row=2, col=1, showgrid=False, zeroline=False)
        
        fig.update_yaxes(title="Yield<br>(% of disease free)", row=1, col=1)
        
        fig.update_yaxes(title="Resistance<br>frequency", row=2, col=1)

        return fig



    def _get_corner_text_labels(self):
        top_row = 1.08
        bottom_row = 0.52
        
        left = 0.02

        cA = get_big_text_annotation(left, top_row, 'A')
        cB = get_big_text_annotation(left, bottom_row, 'B')
        
        out = [cA, cB]
        return out
    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/single/", "/paper_figs/yield_rf_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)







class ParamScanPlotMeVsHobb:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 800

        self.data = self._process_data(data)

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)
    



    def _process_data(self, data):

        req_cols = data[['run', 'RFB_maxContEL', 'EqSel_maxContEL',
                                 "best_value_RFB", "best_value_EqSel",
                                 "RS", "SR",
                                 "RR", "sr_prop",
                                #  "delta_1", "delta_2",
                                 "omega_1", "omega_2",
                                 ]]
        
        complete = req_cols.dropna()

        complete['successMetric'] = self._get_success_metric(complete)

        complete['IRFMetric'] = complete.apply(lambda x: log10(x['RS']) - log10(x['SR']), axis=1)
        
        complete['AsympMetric'] = complete.apply(lambda x: x['omega_1']/(x['omega_1']+x['omega_2']), axis=1)

        return complete



    @staticmethod
    def _get_success_metric(x):

        bv_ES = np.asarray(x['best_value_EqSel'])
        ES_EL = np.asarray(x['EqSel_maxContEL'])
        bv_RFB = np.asarray(x['best_value_RFB'])
        RBF_EL = np.asarray(x['RFB_maxContEL'])

        out = [100*max(bv_ES[i], ES_EL[i]) / max(bv_RFB[i], RBF_EL[i]) for i in range(len(bv_ES))]

        return out



        


    def _generate_figure(self):
        trace_dict = self._get_traces()

        ugly_fig = self._add_traces_to_figure(trace_dict)

        fig = self._sort_layout(ugly_fig)

        return fig



    def _get_traces(self):        
        out = {}

        data = self.data
        
        data['logRR'] = [log10(x) for x in data['RR']]        

        for key in ['IRFMetric', 'AsympMetric', 'logRR', 'sr_prop']:
            
            scatter = go.Scatter(x=data[key],
                y=data['successMetric'],
                mode='markers',
                marker=dict(opacity=0.2))

            out[key] = scatter

        return out




    def _add_traces_to_figure(self, trace_dict):
        fig = make_subplots(rows=2, cols=2, vertical_spacing=0.3)

        fig.add_trace(trace_dict['IRFMetric'], row=1, col=1)
        fig.add_trace(trace_dict['AsympMetric'], row=2, col=1)
        fig.add_trace(trace_dict['logRR'], row=1, col=2)
        fig.add_trace(trace_dict['sr_prop'], row=2, col=2)

        return fig



    def _sort_layout(self, fig):
        fig = self._update_axes(fig)
        fig = self._update_layout(fig)
        fig = self._add_corner_text_labels(fig)
        return fig


    @staticmethod
    def _update_axes(fig):
        fig.update_xaxes(title="Log difference in <br>single resistance frequencies", row=1, col=1)
        fig.update_xaxes(title="Asymptote metric", row=2, col=1)
        fig.update_xaxes(title="Double resistant<br>frequency (log scale)", row=1, col=2)
        fig.update_xaxes(title="Proportion of sex", row=2, col=2)
        
        fig.update_yaxes(title="Success metric", row=1, col=1)
        fig.update_yaxes(title="Success metric", row=2, col=1)

        return fig

    
    def _update_layout(self, fig):
        fig.update_layout(standard_layout(False, self.width, self.height))
        return fig


    def _add_corner_text_labels(self, fig):
        top_row = 1.06
        bottom_row = 0.42
        
        left = -0.01
        middle = 0.56

        c1 = get_big_text_annotation(left, top_row, 'A')
        c2 = get_big_text_annotation(middle, top_row, 'B')
        c3 = get_big_text_annotation(left, bottom_row, 'C')
        c4 = get_big_text_annotation(middle, bottom_row, 'D')
        
        annotz = [c1, c2, c3, c4]

        fig.update_layout(annotations=annotz)
        return fig
    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = f"../outputs/figures/paper_figs/param_scan_hobb_me_{conf_str}"
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)






class ParamScanPlotHighLowDose:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 800

        self.data = self._process_data(data)

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)
    



    @staticmethod
    def _process_data(data):

        req_cols = data[['run', 'RFB_minDS', 'RFB_maxDS', 
                            'max_dose_sums',
                            'min_dose_sums',
                            "RS", "SR", "RR", "sr_prop", "delta_1", "delta_2"]]
        
        complete = req_cols.dropna()

        complete['minMixStrength'] = complete.apply(lambda x: (
            100*(x['RFB_minDS']- x['min_dose_sums'])) / (x['max_dose_sums']
             - x['min_dose_sums'])
            , axis=1)
        
        complete['maxMixStrength'] = complete.apply(lambda x: (
            100*(x['RFB_maxDS']- x['min_dose_sums'])) / (x['max_dose_sums']
             - x['min_dose_sums'])
            , axis=1)
        
        
        complete['meanMixStrength'] = complete.apply(lambda x: (
            100*(x['RFB_minDS'] - x['min_dose_sums'] + 0.5*(x['RFB_maxDS'] - x['RFB_minDS']))) / (x['max_dose_sums']
             - x['min_dose_sums'])
            , axis=1)
        
        print(complete['meanMixStrength'])

        complete['IRFMetric'] = complete.apply(lambda x: log10(x['RS']) - log10(x['SR']), axis=1)
        
        complete['DecayMetric'] = complete.apply(lambda x: x['delta_1']/(x['delta_1']+x['delta_2']), axis=1)

        return complete



        


    def _generate_figure(self):
        trace_dict = self._get_traces()

        ugly_fig = self._add_traces_to_figure(trace_dict)

        fig = self._sort_layout(ugly_fig)

        return fig



    def _get_traces(self):        
        out = {}

        data = self.data
        
        # y = data['minMixStrength']
        # y = data['maxMixStrength']
        y = data['meanMixStrength']
        
        x1 = data['IRFMetric']
        # x1 = [log10(x) for x in data['SR']]

        x2 = data['DecayMetric']

        x3 = [log10(x) for x in data['RR']]
        
        x4 = data['sr_prop']
        
        out['IRFMetric'] = [self._get_scatter(x1, y)]

        out['DecayMetric'] = [self._get_scatter(x2, y)]
        
        out['RR'] = [self._get_scatter(x3, y)]

        out['sr_prop'] = [self._get_scatter(x4, y),
                            self.get_best_fit(x4, y, "blue")
                            ]

        return out




    def _get_scatter(self, x, y):
        return go.Scatter(x=x,
            y=y,
            mode='markers',
            marker=dict(opacity=0.2))
        


    @staticmethod
    def get_best_fit(xx, y, col):
        slope, intercept, _, _, _ = stats.linregress(xx,y)

        line = slope*np.asarray(xx) + intercept
        
        best_fit = go.Scatter(
                    x=xx,
                    y=line,
                    mode='lines',
                    marker=dict(color=col),
                    name='Fit'
                    )

        return best_fit



    def _add_traces_to_figure(self, trace_dict):
        fig = make_subplots(rows=2, cols=2, vertical_spacing=0.3)
        
        self.add_traces_each_subplot(fig, 1, 1, trace_dict['IRFMetric'])
        self.add_traces_each_subplot(fig, 2, 1, trace_dict['DecayMetric'])
        self.add_traces_each_subplot(fig, 1, 2, trace_dict['RR'])
        self.add_traces_each_subplot(fig, 2, 2, trace_dict['sr_prop'])
        
        return fig





    @staticmethod
    def add_traces_each_subplot(fig, row, col, traces):
        for trace in traces:
            fig.add_trace(trace, row=row, col=col)
        return fig

    def _sort_layout(self, fig):
        fig = self._update_axes(fig)
        fig = self._update_layout(fig)
        fig = self._add_corner_text_labels(fig)
        return fig
    


    @staticmethod
    def _update_axes(fig):
        fig.update_xaxes(title="Log difference in <br>single resistance frequencies", row=1, col=1)
        fig.update_xaxes(title="Decay rate metric", row=2, col=1)
        fig.update_xaxes(title="Double resistant<br>frequency (log scale)", row=1, col=2)
        fig.update_xaxes(title="Proportion of sex", row=2, col=2)
        
        fig.update_yaxes(title="Optimal distance<br>along contour (%)", row=1, col=1)
        fig.update_yaxes(title="Optimal distance<br>along contour (%)", row=2, col=1)

        return fig

    
    def _update_layout(self, fig):
        fig.update_layout(standard_layout(False, self.width, self.height))
        return fig


    def _add_corner_text_labels(self, fig):
        top_row = 1.06
        bottom_row = 0.42
        
        left = -0.01
        middle = 0.56

        c1 = get_big_text_annotation(left, top_row, 'A')
        c2 = get_big_text_annotation(middle, top_row, 'B')
        c3 = get_big_text_annotation(left, bottom_row, 'C')
        c4 = get_big_text_annotation(middle, bottom_row, 'D')
        
        annotz = [c1, c2, c3, c4]

        fig.update_layout(annotations=annotz)
        return fig
    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = f"../outputs/figures/paper_figs/param_scan_high_low_{conf_str}"
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)









class FigTemplateClass:
    def __init__(self, data, conf_str) -> None:

        self.width = FULL_PAGE_WIDTH

        self.height = 620

        self.data = data

        fig = self._generate_figure()

        self._save_and_show(fig, conf_str)



    def _generate_figure(self):
        traces = self._get_traces()

        ugly_fig = self._add_traces_to_figure(traces)

        fig = self._sort_layout(ugly_fig)

        return fig



    def _get_traces(self):        
        traces = []

        return traces


    def _add_traces_to_figure(self, traces):
        fig = go.Figure(data=traces, layout=standard_layout(False, self.width, self.height))
        return fig



    def _sort_layout(self, fig):
        fig = self._update_axes(fig)
        return fig
    


    @staticmethod
    def _update_axes(fig):
        fig.update_xaxes(title="x")
        fig.update_yaxes(title="y")
        return fig

    
    
    
    
    def _save_and_show(self, fig, conf_str):
        fig.show()
        filename = conf_str.replace("/grid/", "/paper_figs/doses_scatter_")
        
        print("saving figure to: \n", filename)

        fig.write_image(filename)


from abc import ABC, abstractmethod
import numpy as np
from math import ceil, floor
from scipy.integrate import ode
import pickle
import os
from tqdm import tqdm
import itertools




from model.params import PARAMS

from model.utils import res_prop_calculator, yield_calculator, \
    SelectionFinder, FungicideStrategy, object_dump

from model.ode_system import ODESystem

from model.config_classes import SingleConfig

from model.outputs import SimOutput, SingleTacticOutput, \
    GridTacticOutput





class Simulator(ABC):
    @abstractmethod
    def run(self):
        pass





class SimulatorWithDisease(Simulator):
    """ Simulate a single season """

    def __init__(self, fungicide_params):
        
        self.ode_sys = ODESystem(fungicide_params)
        
        self.sol = ode(self.ode_sys.system).set_integrator('dopri5',
                                                    nsteps=PARAMS.nstepz)

        self.selection_finder = SelectionFinder
        self.res_prop_finder = res_prop_calculator
        self.yield_finder = yield_calculator

        self.times = ModelTimes(PARAMS)
        



    def run(self, fung1_doses, fung2_doses, primary_inoc):

        self.out = SimOutput(self.times.t)

        self.primary_inoc = primary_inoc
        self.fung1_doses = fung1_doses
        self.fung2_doses = fung2_doses

        self._solve_ode()
        
        final_res_dict, end_freqs = self.res_prop_finder(self.out.states)

        selection = self.selection_finder(self.primary_inoc, final_res_dict).sel

        self.out.final_res_vec_dict = final_res_dict
        self.out.end_freqs = end_freqs
        self.out.selection = selection

        self.out.states.delete_unnecessary_vars()

        return self.out





    def _solve_ode(self):
        """
        Solves the ODE in 4 stages:
        
        - before first spray
        - spray 1
        - spray 2
        - yield period (end of season)

        """
        y0_new = None
        
        list_of_tvs = self.times.t_vecs
        segments = self.times.seg_names

        sol = self.sol

        for time_vec, segment in zip(list_of_tvs, segments):

            y0_new = self._get_y0_this_segment(segment, sol)

            y_array = self._solve_for_y_this_segment(y0_new, sol, time_vec)

            if segment=="yield":
                # final segment - need to add final time
                # rather than leave it for start condition 
                # of next segment

                y_array[:,-1] = sol.y
                y_out = y_array

                y_yield = self._get_yield_contributing_y(y_array)
                t_yield = self.times.t_yield

                self.out.yield_val = self.yield_finder(y_yield, t_yield)

            else:
                y_out = y_array[:,:-1]

            self.out.states.update_y(y_out)





    def _get_y0_this_segment(self, segment, sol):
        
        if segment=="start":
            PI = self.primary_inoc

            y0_new = [PARAMS.S_0] + [0]*9 + [PI['RR'], 
                                                PI['RS'],
                                                PI['SR'],
                                                PI['SS']] + [0]*2

        else:
            y0_new = sol.y

            if segment in ["spray_1", "spray_2"]:
                y0_new[PARAMS.fung_1_ind] = y0_new[PARAMS.fung_1_ind] + self.fung1_doses[segment]
                y0_new[PARAMS.fung_2_ind] = y0_new[PARAMS.fung_2_ind] + self.fung2_doses[segment]
        
        return y0_new
    
    
    
    @staticmethod
    def _solve_for_y_this_segment(y0_new, sol, time_vec):
        
        sol.set_initial_value(y0_new, time_vec[0])

        y_array = np.zeros((PARAMS.no_variables, len(time_vec)))

        for index, t in enumerate(time_vec[1:]):
            if sol.successful():
                y_array[:,index] = sol.y
                sol.integrate(t)
            else:
                raise RuntimeError('ode solver unsuccessful')
        
        return y_array
    

    def _get_yield_contributing_y(self, y):
        """ Sum over S, ERR, ERS, ESR, ESS strains """
        return (y[PARAMS.S_ind,:]
                    + y[PARAMS.ERR_ind,:]
                    + y[PARAMS.ERS_ind,:]
                    + y[PARAMS.ESR_ind,:]
                    + y[PARAMS.ESS_ind,:])








class ModelTimes:
    def __init__(self, params) -> None:
        self.params = params

        self.seg_times = [self.params.T_emerge,
                        self.params.T_GS32,
                        self.params.T_GS39,
                        self.params.T_GS61,
                        self.params.T_GS87]

        
        self.seg_names = ["start", "spray_1", "spray_2", "yield"]

        self.t_vecs = self._get_list_of_time_vecs()

        self.t = self._get_t()


    def _get_list_of_time_vecs(self):
        
        seg_ts = self.seg_times
        
        sum_ns = 0

        list_of_tvs = []

        for ii, segment in enumerate(self.seg_names):

            if segment=="yield":
                # makes sure total number of points is self.params.t_points
                n = 3 + self.params.t_points - sum_ns

            else:
                # make n so that values are approx self.params.dt apart
                n = 1 + (seg_ts[ii+1]-seg_ts[ii])/self.params.dt
                n = floor(n)
                sum_ns += n

            time_vec = np.linspace(seg_ts[ii], seg_ts[ii+1], n)

            if segment=="yield":
                self.t_yield = time_vec

            list_of_tvs.append(time_vec)

        return list_of_tvs


    def _get_t(self):
        tvs = self.t_vecs

        out = np.concatenate([tvs[ii][:-1] if ii!=3 else tvs[ii]
                                for ii in range(len(tvs))])
        return out









class SimulatorDiseaseFree(Simulator):
    """ Simulates a single season, but only returns the yield """

    def __init__(self, fungicide_params):
        ode_sys = ODESystem(fungicide_params)
        self.yield_finder = yield_calculator
        self.sol = ode(ode_sys.system).set_integrator('dopri5', nsteps=PARAMS.nstepz)
        



    def run(self):
        """ Get yield for a single dis-free season """

        sol = self.sol
        
        y0 = [PARAMS.S_0] + [0]*(PARAMS.no_variables-1)
        
        t_not_yield, t_yield = self._get_dis_free_t_vecs()

        _ = self._solve_for_y_this_segment(y0, sol, t_not_yield)

        y_yield = self._solve_for_y_this_segment(sol.y, sol, t_yield)

        y_yield[:, -1] = sol.y

        dis_free_yield = yield_calculator(y_yield[0,:], t_yield)

        return dis_free_yield



    @staticmethod
    def _get_dis_free_t_vecs():
        t0 = PARAMS.T_emerge
        t1 = PARAMS.T_GS61
        t2 = PARAMS.T_GS87
        
        n1 = 1 + (t1-t0)/PARAMS.dt
        n2 = 1 + (t2-t1)/PARAMS.dt
        
        c1 = ceil(n1-0.5)
        c2 = ceil(n2-0.5)
        
        t_not_yield = np.linspace(t0,t1,c1)
        t_yield = np.linspace(t1,t2,c2)
        
        return t_not_yield, t_yield
   
    
    
    @staticmethod
    def _solve_for_y_this_segment(y0_new, sol, time_vec):
        
        sol.set_initial_value(y0_new, time_vec[0])

        y_array  = np.zeros((PARAMS.no_variables, len(time_vec)))

        for index, t in enumerate(time_vec[1:]):
            if sol.successful():
                y_array[:,index] = sol.y
                sol.integrate(t)
            else:
                raise RuntimeError('ode solver unsuccessful')
        
        return y_array







# * End of Sim cls

class RunModel(ABC):
    @abstractmethod
    def run(self):
        pass
    
    @abstractmethod
    def _load(self):
        pass
    
    @abstractmethod
    def _save(self):
        pass






class RunSingleTactic(RunModel):
    def __init__(self, fcide_parms=None):

        self.sim = SimulatorWithDisease(fcide_parms)
        
        df_sim = SimulatorDiseaseFree(fcide_parms)

        self.dis_free_yield = df_sim.run()

        self.yield_stopper = 95

        self.PATHOGEN_STRAIN_NAMES = ['RR', 'RS', 'SR', 'SS']




    def run(self, conf):
        """Run HRHR model for single tactic."""
        
        self.filename = conf.config_string

        if conf.load_saved:
            loaded_run = self._load()
            if loaded_run is not None:
                return loaded_run


        self.n_years = len(conf.fung1_doses['spray_1'])

        self.out = SingleTacticOutput(PARAMS.yield_threshold, 
                                            conf.res_props,
                                            self.PATHOGEN_STRAIN_NAMES,
                                            self.n_years, 
                                            self.dis_free_yield)

        self._set_first_year_start_freqs(conf)

        self._loop_over_years(conf)

        self.out.delete_unnecessary_vars()
        
        if conf.save:
            self._save()
        
        return self.out

    





    def _set_first_year_start_freqs(self, conf):
        primary_inoculum = conf.primary_inoculum
        
        if primary_inoculum is None:
            primary_inoculum = self._primary_calculator(conf,
                                                conf.res_props['f1'], 
                                                conf.res_props['f2'])

        self.out.update_start_freqs(primary_inoculum, 0)


    
    def _loop_over_years(self, conf):
        for yr in range(self.n_years):
            # stop the solver after we drop below threshold
            if not (yr>0 and self.out.yield_vec[yr-1]<self.yield_stopper):
                self._run_single_year(conf, yr)
        
        if min(self.out.yield_vec)>PARAMS.yield_threshold:
            self.out.failure_year = -1




    def _run_single_year(self, conf, yr):
        
        fung1_doses = self._get_dose_this_fung_this_yr(conf.fung1_doses, yr)
        fung2_doses = self._get_dose_this_fung_this_yr(conf.fung2_doses, yr)
        
        model_inoc_in = self._get_initial_density(yr)
        
        sim_out = self.sim.run(fung1_doses, fung2_doses, model_inoc_in)

        self.out.add_new_sim_output(sim_out, yr)

        self._set_next_year_start_freqs(conf, sim_out, yr+1)
    



    def _set_next_year_start_freqs(self, conf, sim_out, next_yr):

        freqs_out = sim_out.end_freqs

        # sex/asex after each season
        res_prop_1_end = freqs_out['RR'] + freqs_out['RS']
        res_prop_2_end = freqs_out['RR'] + freqs_out['SR']

        # get next year's primary inoc - including SR step
        next_year_vals = self._primary_calculator(conf,
                                                    res_prop_1_end,
                                                    res_prop_2_end,
                                                    freqs_out)

        self.out.update_start_freqs(next_year_vals, next_yr)







    @staticmethod
    def _get_dose_this_fung_this_yr(dose_vec, yr):
        return dict(spray_1 = dose_vec['spray_1'][yr],
                    spray_2 = dose_vec['spray_2'][yr])



    def _get_initial_density(self, yr):
        out = {}
        for key in self.PATHOGEN_STRAIN_NAMES:
            out[key] = PARAMS.init_den*self.out.start_freqs[key][yr]
        return out

    

    @staticmethod
    def _primary_calculator(
                conf,
                res_prop_1,
                res_prop_2,
                proportions=None,
                ):
        
        # is_mixed_sex = conf.is_mixed_sex
        sex_prop = conf.sex_prop

        sex = dict(
            RR = res_prop_1*res_prop_2,
            RS = res_prop_1*(1-res_prop_2),
            SR = (1-res_prop_1)*res_prop_2,
            SS = (1-res_prop_1)*(1-res_prop_2)
            )

        if proportions is None:
            return sex

        else:
            asex = proportions
            out = {}
            for key in sex.keys():
                out[key] = sex_prop*sex[key] + (1 - sex_prop)*asex[key]
            
            return out
    




    # load/save single

    def _load(self):
        filename = self.filename
        
        if os.path.isfile(filename) and "single" in filename:
            loaded_run = pickle.load(open(filename, 'rb'))
            return loaded_run
        else:
            return None


    def _save(self):
        if "single" in self.filename:
            object_dump(self.filename, self.out)



# * End of RunSingleTactic





















class RunGrid(RunModel):
    def __init__(self, fcide_parms=None):
        self.sing_tact = RunSingleTactic(fcide_parms)
        self.fung_strat = FungicideStrategy



    def run(self, conf):

        self.filename = conf.config_string
        
        if conf.load_saved:
            loaded_run = self._load()
            if loaded_run is not None:
                return loaded_run

        self.output = GridTacticOutput(conf.n_doses, conf.n_years)

        self._run_the_grid(conf)

        if conf.save:
            self._save()

        return self.output



    def _run_the_grid(self, conf):
        fs = self.fung_strat(conf.strategy, conf.n_years)
        
        for f1_ind in tqdm(range(conf.n_doses)):
            for f2_ind in range(conf.n_doses):

                conf.fung1_doses, conf.fung2_doses = fs.get_grid_doses(f1_ind,
                                                        f2_ind, conf.n_doses)

                one_tact_output =  self.sing_tact.run(conf)
                
                self._post_process(one_tact_output, f1_ind, f2_ind)





    
    def _post_process(self, data_this_dose, f1_ind, f2_ind):

        self.output.LTY[f1_ind,f2_ind] = self._lifetime_yield(data_this_dose.yield_vec, data_this_dose.failure_year)

        self.output.TY[f1_ind,f2_ind] = self._total_yield(data_this_dose.yield_vec)

        self.output.FY[f1_ind,f2_ind] = data_this_dose.failure_year

        self.output.yield_array[f1_ind,f2_ind,:] = data_this_dose.yield_vec


        self.output.update_dicts_of_arrays(data_this_dose, f1_ind, f2_ind)

        



    @staticmethod
    def _lifetime_yield(Y_vec, F_y):
        return sum(Y_vec[:(F_y+1)])/100




    @staticmethod
    def _total_yield(Y_vec):
        return sum(Y_vec)/100




    # load/save single

    def _load(self):
        filename = self.filename

        if os.path.isfile(filename):
            loaded_run = pickle.load(open(filename, 'rb'))
            return loaded_run
        else:
            return None



    def _save(self):
        object_dump(self.filename, self.output)




# End of RunGrid cls










# # * changing doses fns

def get_SR_by_doses(doses, freqs):
    outputs = {}
    for dose, rf in itertools.product(doses, freqs):
        conf_single = SingleConfig(1, rf, rf, dose, dose, dose, dose)
        output = RunSingleTactic().run(conf_single)
        outputs[f"dose={dose},rf={rf}"] = output


    conf_str = conf_single.config_string_img
    str_freqs = [str(round(f,2)) for f in freqs]
    str_doses = [str(round(d,2)) for d in doses]

    middle_string = ("=" + ",_".join(str_freqs) +
                "_doses=" + ",_".join(str_doses))
    middle_string = middle_string.replace(".", ",")

    conf_str = ("=".join(conf_str.split("=")[0:2]) + 
            middle_string + conf_str.split("=")[-1])

    
    return outputs, conf_str

# * End of changing doses fns



def get_conf_string(folder, filename):
    conf_str = folder + filename.replace(".", ",") + ".pickle"
    config_string = conf_str

    conf_str2 = conf_str.replace("saved_runs", "figures")
    config_string_img = conf_str2.replace("pickle", "png")

    return config_string, config_string_img


class BaselineConfig:
    def __init__(self, n_years, rp1, rp2, primary_inoculum):

        self.load_saved = True

        self.save = True

        self.folder_save_run = '../outputs/saved_runs/'

        self.sex_prop = 0
        
        self.n_years = n_years

        self.res_props = dict(
            f1 = rp1,
            f2 = rp2
            )

        self.primary_inoculum = primary_inoculum



    def add_baseline_str(self):

        rp1 = self.res_props['f1']
        rp2 = self.res_props['f2']

        if self.primary_inoculum is None:
            inoc_str = "N"
        else:
            inoc_str = "Y"

        self.save_string = f"Ny={self.n_years}_" + \
            f"Rps={rp1},_{rp2}_" + \
            f"PI={inoc_str}_" + \
            f"Sex={self.sex_prop}"



class SingleConfig(BaselineConfig):
    def __init__(self,
                n_years,
                rp1,
                rp2,
                d11,
                d12,
                d21,
                d22,
                primary_inoculum=None
                # zeroth_season_reproduction=True
                ):
        
        super().__init__(n_years, rp1, rp2, primary_inoculum)
        
        self.fung1_doses = dict(
            spray_1 = [d11]*n_years,
            spray_2 = [d12]*n_years
            )

        self.fung2_doses = dict(
            spray_1 = [d21]*n_years,
            spray_2 = [d22]*n_years
            )
        
        self.add_string()



    def add_string(self, extra_detail=None):
        d11 = round(self.fung1_doses['spray_1'][0],6)
        d12 = round(self.fung1_doses['spray_2'][0],2)
        d21 = round(self.fung2_doses['spray_1'][0],6)
        d22 = round(self.fung2_doses['spray_2'][0],2)

        self.add_baseline_str()
        filename = f"single/{self.save_string}_doses={d11},{d12},{d21},{d22}"
        if extra_detail is not None:
            filename = filename + extra_detail

        self.config_string, self.config_string_img = get_conf_string(self.folder_save_run, filename)




class GridConfig(BaselineConfig):
    def __init__(self,
            n_years,
            rp1,
            rp2,
            n_doses,
            primary_inoculum=None
            # zeroth_season_reproduction=True
            ):

        super().__init__(n_years, rp1, rp2, primary_inoculum)
        
        self.strategy = 'mix'

        self.n_doses = n_doses

        self.add_string()


    def add_string(self, extra_detail=None):
        self.add_baseline_str()
        filename = f"grid/{self.save_string}_Nd={self.n_doses}_S={self.strategy}"
        
        if extra_detail is not None:
            filename = filename + extra_detail
        
        self.config_string, self.config_string_img = get_conf_string(self.folder_save_run, filename)
        



import numpy as np
import pandas as pd

from model.config_classes import GridConfig, SingleConfig
from model.simulator import RunSingleTactic





class RandomPars:
    """
    Generates random parameters for model run
    
    Can return
    - pars
    - grid config
    - single config
    """

    def __init__(self, config, run_index) -> None:

        self.config = config

        self._run_index = run_index



    def find_pars(self):
        """
        returns rfs1, rfs2, rfD, om_1, om_2, delt_1, delt_2, sr_prop
        """

        conf = self.config

        pars_are_valid = False

        while pars_are_valid is False:
        
            rfs1_power = np.random.uniform(low=conf["RFS1"][0], high=conf["RFS1"][1])
            rfs2_power = np.random.uniform(low=conf["RFS2"][0], high=conf["RFS2"][1]) 
            rfD_power = np.random.uniform(low=conf["RFD"][0], high=conf["RFD"][1])
            
            rfs1 = 10**(rfs1_power)
            rfs2 = 10**(rfs2_power)
            rfD = 10**(rfD_power)

            om_1 = np.random.uniform(low=conf["asym1"][0], high=conf["asym1"][1])
            om_2 = np.random.uniform(low=conf["asym2"][0], high=conf["asym2"][1])
            
            delt_1 = np.random.uniform(low=conf["dec_rate1"][0], high=conf["dec_rate1"][1])
            delt_2 = np.random.uniform(low=conf["dec_rate2"][0], high=conf["dec_rate2"][1])
            
            theta_val = conf["theta"]
            
            sr_prop = np.random.uniform(low=conf["SR"][0], high=conf["SR"][1])

            fungicide_params = self.get_fung_parms_dict(om_1, om_2, delt_1, delt_2, theta_val)

            pathogen_pars = dict(rfs1=rfs1,
                rfs2=rfs2,
                rfD=rfD, 
                sr_prop=sr_prop)

            pars_are_valid = self._check_validity(fungicide_params, pathogen_pars)

        
        self.get_inoc_dict(rfs1, rfs2, rfD)
        
        self.fung_parms = fungicide_params

        self.path_and_fung_pars = rfs1, rfs2, rfD, om_1, om_2, delt_1, delt_2

        self.sr_prop = sr_prop








    def _check_validity(self, fungicide_params, pathogen_pars):

        self.get_inoc_dict(pathogen_pars['rfs1'],
                    pathogen_pars['rfs2'],
                    pathogen_pars['rfD'])
        
        
        corners = [[1,0], [0,1], [1,1]]

        for ii, jj in corners:
            this_dose_yield = self._get_yield_these_doses(fungicide_params,
                                                pathogen_pars, ii, jj)

            if this_dose_yield<=95:
                return False
        
        return True





    def _get_yield_these_doses(self, fungicide_params, pathogen_pars, d1, d2):

        ConfigSingleRun = SingleConfig(1, None, None, d1, d1, d2, d2, primary_inoculum=self.inoc)
        
        ConfigSingleRun.sex_prop = pathogen_pars['sr_prop']

        ConfigSingleRun.load_saved = False

        this_run = RunSingleTactic(fungicide_params).run(ConfigSingleRun)
        
        yield_out = this_run.yield_vec[0]

        return yield_out





    def get_inoc_dict(self, rfs1, rfs2, rfD):
        self.inoc = dict(RR = rfD,
                         RS = rfs1,
                         SR = rfs2,
                         SS = 1 - rfD - rfs1 - rfs2)

    

    def get_fung_parms_dict(self, omega_1, omega_2, delta_1, delta_2, theta):
        return dict(omega_1 = omega_1,
                    omega_2 = omega_2,
                    theta_1 = theta,
                    theta_2 = theta,
                    delta_1 = delta_1,
                    delta_2 = delta_2)


    
    def get_all_parms_df(self):
        out = {**self.fung_parms,
                **self.inoc,
                "sr_prop": self.sr_prop,
                "run": self._run_index}
        
        self.par_df = pd.DataFrame([out])





    # single and grid config finders:

    def get_grid_conf(self, n_doses):
        
        conf = GridConfig(self.config['n_years'], None, None, n_doses,
                                    primary_inoculum=self.inoc)

        config_out = self._process_conf(conf)

        self.grid_conf = config_out

        return config_out





    def get_single_conf(self, dose1, dose2):

        conf = SingleConfig(self.config['n_years'], None, None,
                                dose1, dose1, dose2, dose2,
                                primary_inoculum=self.inoc)
        
        config_out = self._process_conf(conf)

        self.sing_conf = config_out

        return config_out




    def _process_conf(self, conf):

        conf.sex_prop = self.sr_prop

        conf.load_saved = self.config['load_saved']
        
        conf.save = self.config['save']

        conf.add_string()

        config_out = self._update_par_scan_conf_str(conf)

        return config_out



    
    def _update_par_scan_conf_str(self, conf):
        
        rfs1, rfs2, rfD, om_1, om_2, delt_1, delt_2 = self.path_and_fung_pars

        conf_str = conf.config_string_img
        conf_str = conf_str.replace("grid", "param_scan")

        par_str = (f"_fung_pars={round(om_1,6)},{round(om_2,6)}," + 
                    f"{round(delt_1,6)},{round(delt_2,6)}," +
                    f"rf1={round(rfs1,10)},rf2={round(rfs2,10)},rfd={round(rfD,15)}")
        
        par_str = par_str.replace(".", ",")
        conf_str = conf_str.replace(".png", par_str + ".png")
        
        conf.config_string_img = conf_str
        
        saved_run_str = conf_str.replace(".png", ".pickle")
        
        conf.config_string = saved_run_str.replace("figures/", "saved_runs/")

        return conf






import pandas as pd
import copy
import numpy as np


from model.simulator import RunGrid
from plotting.figures import DoseSpaceScenariosPlot
from .pars import RandomPars

# TOC
# combine_PS_rand_outputs
# PostProcess
# MaxAlongContourDF



def combine_PS_rand_outputs(config, seeds):

    df = pd.DataFrame()
    
    par_str = config['par_str']

    folder = config['folder_save']

    for seed in seeds:

        temporary = pd.read_csv(f"{folder}/par_scan/summary_df_seed={seed}_{par_str}.csv")

        df = df.append(temporary, ignore_index=True)

    
    combined_filename = f"{folder}/combined/output_summary_{par_str}.csv"
    print(f"saving combined output to {combined_filename}")
    df.to_csv(combined_filename)





class PostProcess:

    def __init__(self, folder, par_str):
        df_in = pd.read_csv(f"{folder}/combined/output_summary_{par_str}.csv")
        self.df = df_in.drop(["Unnamed: 0"], axis=1)

        self.folder = folder
        self.par_str = par_str
        







    def get_maximum_along_contour_df(self):
        self.max_along_contour_df = MaxAlongContourDF(self.folder, self.df).df


    











    def analyse_max_contour_df(self):

        df = copy.copy(self.max_along_contour_df)

        # df = df[df['min_corner']>0]

        eq_sel_df = self._get_non_null_df(df, "EqSel_worked_geq")

        self.filtered_dataframe_outcome(df, "RFB_maxCont%")
        self.filtered_dataframe_outcome(df, "fullDose%")
        self.filtered_dataframe_outcome(df, "minEqDose%")

        self.filtered_dataframe_outcome(eq_sel_df, "EqSel_maxCont%")
        self.filtered_dataframe_outcome(eq_sel_df, "EqSel_lowDoseMax%")

        self.check_IVT_method(df, "RFB")
        self.check_IVT_method(eq_sel_df, "EqSel")





    def _get_non_null_df(self, df_in, col_to_check_if_null):
        invalid_runs = df_in[col_to_check_if_null].isnull()
        return df_in[~invalid_runs]



    def filtered_dataframe_outcome(self, df, strategy):

        mean = df[strategy].mean()

        conditional_mean = df[df[strategy]<100][strategy].mean()

        worked = df[df[strategy]>=100].shape[0]
        greater = df[df[strategy]>100].shape[0]
        lesser = df[df[strategy]<100].shape[0]
        equal = df[df[strategy]==100].shape[0]

        total = df.shape[0]

        sum_total = sum([greater, lesser, equal])

        out = dict(worked=worked, 
                    greater=greater,
                    lesser=lesser,
                    equal=equal,
                    total=total,
                    sum_total=sum_total,
                    work_pc=round(100*worked/sum_total,1),
                    mean=round(mean,1),
                    conditional_mean=round(conditional_mean,1),
                    strategy=strategy
                    )
        
        print(out)

        return out



    def check_IVT_method(self, df, method):

        # if False: # method=="EqSel":
        #     IVT_true = df[f'best_region_{method}']=="True"
        #     IVT_false = df[f'best_region_{method}']=="False"
        #     IVT_NA = df[f'best_region_{method}'].isin(["True", "False"])
        # else:
        
        IVT_true = df[f'best_region_{method}']==True
        IVT_false = df[f'best_region_{method}']==False
        IVT_NA = df[f'best_region_{method}'].isin([True, False])

        opt_true = (df[f'{method}_worked_geq']==True)
        opt_false = (df[f'{method}_worked_geq']==False)
        opt_NA = (df[f'{method}_worked_geq'].isnull())

        out = dict(
            IVT_method_T = df[IVT_true].shape[0],
            IVT_method_F = df[IVT_false].shape[0],
            IVT_method_NA = df[~IVT_NA].shape[0],
            
            opt_method_T = df[opt_true].shape[0],
            opt_method_F = df[opt_false].shape[0],
            opt_method_NA = df[opt_NA].shape[0],

            both_succeeded = df[(opt_true & IVT_true)].shape[0],
            both_failed = df[(~opt_true & ~IVT_true)].shape[0],
            
            either_succeeded = df[(opt_true | IVT_true)].shape[0],
            opt_but_not_IVT = df[(opt_true & ~IVT_true)].shape[0],
            IVT_but_not_opt = df[(~opt_true & IVT_true)].shape[0],
            method = method,
            )

        failed = df[((~opt_true) & (~IVT_true))]
        
        failed_runs = failed[[f'max_grid_EL', f'{method}_maxContEL', f'best_value_{method}', 
                    f'{method}_diff_from_opt', 'run']].sort_values(by=[f'{method}_diff_from_opt', 'run'])

        print("\n")
        print(out)
        print("\n")
        print(f"Testing {method}; these runs failed on both methods:")
        print("\n")        
        
        print(failed_runs)






    def analyse_failed(self):

        df = copy.copy(self.max_along_contour_df)

        fail = self._get_failed_runs(df)

        print("\n")
        print("These runs failed:\n")

        print(fail[['RFB_diff_from_opt',
                    'run',
                    'max_grid_EL',
                    'RFB_maxContEL',
                    'best_value_RFB',
                    ]].to_string())

        
    @staticmethod
    def _get_failed_runs(df):
        return df[df['RFB_maxCont%']<100]

    



    def which_runs_worked_max_cont(self):
        df = copy.copy(self.max_along_contour_df)

        failed = df[df['RFB_maxCont%']<100]

        runs_that_failed = failed["run"].unique()

        failed_pars = self.get_params_for_specific_runs(runs_that_failed)

        n_fail = failed_pars.shape[0]

        failed_pars.to_csv(f"{self.folder}/par_scan/failed_{n_fail}.csv")

    





    def get_params_for_specific_runs(self, which_runs):

        par_df = copy.copy(self.df)

        out = pd.DataFrame()
        for rr in which_runs:
            this_run = par_df[par_df["run"]==rr].iloc[0,:]
            out = out.append(this_run, ignore_index=True)
        
        return out








        




    def check_high_or_low_dose(self):

        my_df = copy.copy(self.df)

        my_df['high_better_than_low'] = my_df['RFB_highDoseMaxEL'] >= my_df['RFB_lowDoseMaxEL']

        # strats = ["minEqDose", "fullDose"]
        
        # for string in strats:
        #     my_df[string + "%"] = 100*my_df[string + "EL"]/my_df["max_grid_EL"]
        
        grouped = my_df.groupby(["run"]).first()

        df_out = pd.DataFrame(grouped)

        df_out = df_out.reset_index()
        
        df_out = df_out.sort_values(['high_better_than_low', 'sr_prop'])

        
        sex_and_high_eff = df_out[((df_out['sr_prop']>0.9)
                            & (df_out['omega_1']>0.9) 
                            & (df_out['omega_2']>0.9) 
                            # & (df_out['delta_1']>0.01)
                            # & (df_out['delta_2']>0.01)  
                            # & (df_out['RS']<0.00001)
                            # & (df_out['SR']<0.00001)
                            )]
        
        # sex_and_high_eff = df_out[(df_out['sr_prop']>0.6)]
        
        print("\n")
        print("worked/total:",sex_and_high_eff['high_better_than_low'].sum(), sex_and_high_eff.shape[0])
        print("\n")
        print(sex_and_high_eff[['high_better_than_low', 'sr_prop',
                'omega_1', 'omega_2', 'delta_1', 'delta_2',
                'RR'
                ]].sort_values(['high_better_than_low', 'sr_prop']))


        # print(df_out.loc[~df_out['high_better_than_low']]['sr_prop'].mean())
                
        filename = f"{self.folder}/par_scan/high_or_low_dose_{len(df_out)}.csv"

        print(f"Saving high or low dose csv to: \n{filename}")
        
        df_out.to_csv(filename)




  
    
    
    def re_run(self, NDoses, run_indices):

        df_test = self.get_params_for_specific_runs(run_indices)

        for ii in range(df_test.shape[0]):

            pars = df_test.iloc[int(ii),:]
           
            print("\nRe-running run:", df_test.iloc[int(ii),:].run, "\n")

            grid_config, fung_params = self._get_grid_config_and_fung_pars(pars, NDoses)

            grid_output = RunGrid(fung_params).run(grid_config)

            conf_str = grid_config.config_string_img

            FY = grid_output['FY']
            opt_region = FY == np.amax(FY)
                
            n_opt_doses = opt_region.sum()

            print(f"Number of optimal dose combos: {n_opt_doses}")

            # plot output
            # dose_grid_heatmap(grid_output, grid_config, "FY", conf_str)
            
            # eq_RFB_contours(grid_output, grid_config, title=f"Run={str(pars.run)}")
            DoseSpaceScenariosPlot(grid_output, conf_str)






    @staticmethod
    def _get_grid_config_and_fung_pars(pars, NDoses):

        config = {'load_saved': True, 'n_years': 35}

        RP = RandomPars(config, None)

        RP.get_inoc_dict(pars["RS"], pars["SR"], pars["RR"])
        
        fung_params = RP.get_fung_parms_dict(pars["omega_1"], pars["omega_2"], 
                                pars["delta_1"], pars["delta_2"])
        
        RP.sr_prop = pars["sr_prop"]

        RP.path_and_fung_pars = (pars["RS"], pars["SR"], pars["RR"], pars["omega_1"],
                    pars["omega_2"], pars["delta_1"], pars["delta_2"])

        grid_config = RP.get_grid_conf(NDoses)

        return grid_config, fung_params














class MaxAlongContourDF:

    def __init__(self, folder, df_input):
        self.folder = folder
        self.get_and_save(df_input)
    


    def get_and_save(self, df_input):
        df_inter = self._get_intermediate_df(df_input)

        self.df = self._tidy_df(df_inter)
        
        self._save_df()





    def _get_intermediate_df(self, data):

        data.fillna(0)

        data = pd.DataFrame(data)
        
        data['maxAlongContour'] = data['RFB_maxContEL'] >= data['max_grid_EL']

        strats = ["RFB_maxCont", "EqSel_maxCont", "minEqDose", "fullDose", "EqSel_lowDoseMax"]
        
        for string in strats:
            data.loc[:, string + "%"] = 100*data[string + "EL"]/data["max_grid_EL"]
            
        data['min_corner'] = data[["corner_01", "corner_10"]].min(axis=1)

        data['RFB_diff_from_opt'] = data.apply(self.get_diff_from_opt_RFB, axis=1)

        data['EqSel_diff_from_opt'] = data.apply(self.get_diff_from_opt_EqSel, axis=1)
        
        return data


    def get_diff_from_opt_RFB(self, data):
        return data['max_grid_EL'] - max(data['RFB_maxContEL'], data['best_value_RFB'])
    
    def get_diff_from_opt_EqSel(self, data):
        return data['max_grid_EL'] - max(data['EqSel_maxContEL'], data['best_value_EqSel'])
    



    def _tidy_df(self, df):

        # avoid the "-1" case where has never failed:
        if df[df["fullDoseEL"]<=0].shape[0]:
            n_never_fail = df[df["fullDoseEL"]<=0].shape[0]
            print(f"{n_never_fail} runs never failed - need longer n_years. Filtering them out for now.")
            df = df[df["fullDoseEL"]>0]
        
        return df





    def _save_df(self):
        df_out = self.df
        
        filename = f"{self.folder}/par_scan/max_along_contour_{len(df_out)}.csv"

        print(f"Saving maximum along contour csv to: \n{filename}")

        df_out.to_csv(filename)





from .config_classes import SingleConfig, GridConfig

n_years = 20

res_f1 = 10**(-5)
res_f2 = 10**(-5)

d11 = 1
d12 = 1
d21 = 0.5
d22 = 0.5


n_grid = 4

ConfigSingleRun = SingleConfig(n_years, res_f1, res_f2, d11, d12, d21, d22)
ConfigGridRun   =   GridConfig(n_years, res_f1, res_f2, n_grid)


import numpy as np

from model.utils import get_SR_by_doses
from plotting.figures import SR_by_dose_plot


SR_by_dose = True

# setup

doses = np.linspace(0, 1, 10)
freqs = [10**(-5), 0.01, 0.02, 0.05, 0.1]


# plot

if SR_by_dose:
    outputs, conf_str = get_SR_by_doses(doses, freqs)
    SR_by_dose_plot(outputs, conf_str)




from model.simulator import RunSingleTactic, RunGrid

from model.config_classes import SingleConfig, GridConfig

from plotting.figures import DiseaseProgressCurvesAll, DoseSpaceScenariosPlot, \
    DosesScatterPlot, YieldAndRfPlot, ParamScanPlotMeVsHobb, \
    ParamScanPlotHighLowDose, CombinedModelPlot

from param_scan.fns.config import config_rand
from param_scan.fns.post_process import PostProcess



# which plots

model_output_overview = False
rf_yield = False
model_output_combined = False
dose_space = True
doses_scatter = False
param_scan_hobb_vs_me = False
param_scan_high_low_dose = False



def get_param_data(par_str):
    PP = PostProcess(par_str)
    PP.get_maximum_along_contour_df()
    data = PP.max_along_contour_df
    return data


# plot

if model_output_overview:
    config_sing = SingleConfig(1, 2*10**(-1), 5*10**(-2), 1, 1, 0.5, 0.5)
    config_sing.load_saved = False
    output = RunSingleTactic().run(config_sing)
    DiseaseProgressCurvesAll(output, config_sing.config_string_img)


if rf_yield:
    config_sing = SingleConfig(10, 10**(-3), 10**(-5), 1, 1, 0.5, 0.5)
    # config_sing.load_saved = False
    run_s = RunSingleTactic()
    run_s.yield_stopper = 0
    output = run_s.run(config_sing)
    YieldAndRfPlot(output, config_sing.config_string_img)


if model_output_combined:
    config_sing = SingleConfig(10, 10**(-3), 10**(-6), 1, 1, 0.5, 0.5)
    config_sing.load_saved = False
    run_s = RunSingleTactic()
    run_s.yield_stopper = 0
    output = run_s.run(config_sing)
    CombinedModelPlot(output, config_sing.config_string_img)


if dose_space:
    ConfigGridRun = GridConfig(30, 10**(-7), 10**(-3), 51)

    ConfigGridRun = GridConfig(30, 10**(-7), 10**(-3), 6)
    ConfigGridRun.load_saved = False
    output = RunGrid().run(ConfigGridRun)
    
    DoseSpaceScenariosPlot(output, ConfigGridRun.config_string_img)


if doses_scatter:
    ConfigGridRun = GridConfig(30, 10**(-7), 10**(-3), 51)    
    output = RunGrid().run(ConfigGridRun)
    DosesScatterPlot(output, ConfigGridRun.config_string_img)


if param_scan_hobb_vs_me:
    par_str = config_rand['par_str']
    data = get_param_data(par_str)
    ParamScanPlotMeVsHobb(data, f"{par_str}.png")


if param_scan_high_low_dose:
    par_str = config_rand['par_str']
    data = get_param_data(par_str)
    ParamScanPlotHighLowDose(data, f"{par_str}.png")



from math import ceil

# alpha: scale factor for omega (max effect)
# of fungicides after resistance (partial res type 1),
# or for curvature (partial res type 2)


class Parameters:
    def __init__(self):
        self.omega_1 = 1
        self.omega_2 = 1
        
        # partial resistance on 
        # - asymptote (PR type 1)
        # - curvature (PR type 2)
        
        # self.alpha_1 = 0
        # self.alpha_1_C = 1
        # self.alpha_2 = 0
        # self.alpha_2_C = 1
        
        self.theta_1 = 9.6
        self.theta_2 = 9.6
        
        # effect on latent period different?
        # self.omega_1_L = 1
        # self.omega_2_L = 1
        
        # self.theta_1_L = legal_dose*t1
        # self.theta_2_L = legal_dose*t2

        self.r = 1.26*10**(-2)
        self.k = 1
        
        self.S_0 = 0.05/4.2
        self.beta = 1.56*10**(-2)
        self.gamma = 1/266
        
        self.mu = 1/456
        self.nu = 8.5*10**(-3)
        
        self.delta_1 = 1.11*10**(-2)
        self.delta_2 = 1.11*10**(-2)
        
        self.init_den = 1.09*10**(-2)/4.2
        
        
        self.T_emerge = 1212
        self.T_GS32 = 1456
        self.T_GS39 = 1700
        self.T_GS61 = 2066
        self.T_GS87 = 2900
        
        self.nstepz = 10**3
        self.dt = 20

        self.t_points = ceil((self.T_GS87 - self.T_emerge)/self.dt)
        
        self.yield_threshold = 95
        
        # self.JSON_path = 'HR_HR/Asexual_config/JSON/'
        # self.pickle_path = 'HR_HR/Asexual_output/Saved_pickles/Cluster_version/'
        
        self.no_variables = 16
        
        self.S_ind = 0
        self.ERR_ind = 1
        self.ERS_ind = 2
        self.ESR_ind = 3
        self.ESS_ind = 4
        
        # self.IRR_ind = 5
        # self.IRS_ind = 6
        # self.ISR_ind = 7
        # self.ISS_ind = 8
        
        # self.R_ind = 9
        
        # self.PRR_ind = 10
        # self.PRS_ind = 11
        # self.PSR_ind = 12
        # self.PSS_ind = 13

        self.fung_1_ind = 14
        self.fung_2_ind = 15


        
        
        
        # kill
        self.IRR_ind = 5
        self.IRS_ind = 6
        self.ISR_ind = 7
        self.ISS_ind = 8
        
        self.R_ind = 9
        
        self.PRR_ind = 10
        self.PRS_ind = 11
        self.PSR_ind = 12
        self.PSS_ind = 13

        self.ER_ind = 1
        self.ES_ind = 4

        self.IR_ind = 5
        self.IS_ind = 8

        self.Fung1_ind = 14
        self.Fung2_ind = 15

        self.res_prop_calc_method = 'final_value'


PARAMS = Parameters()



from model.params import PARAMS
import plotly.colors as pltly_clrs

# ATTRS_DICT = {
#     str(PARAMS.S_ind): dict(name='Susceptible', colour='green', dash="solid", var="S"),
    
#     str(PARAMS.ERR_ind): dict(name='Latent (RR)', colour='rgb(0,0,150)', dash="dot", var="ER"),
#     str(PARAMS.ERS_ind): dict(name='Latent (RS)', colour='rgb(0,0,180)', dash="dash", var="ERS"),
#     str(PARAMS.ESR_ind): dict(name='Latent (SR)', colour='rgb(0,0,210)', dash="dashdot", var="ESR"),
#     str(PARAMS.ESS_ind): dict(name='Latent (SS)', colour='rgb(0,0,240)', dash="solid", var="ES"),
    
#     str(PARAMS.IRR_ind): dict(name='Infected (RR)', colour='rgb(150,0,0)', dash="dot", var="IR"),
#     str(PARAMS.IRS_ind): dict(name='Infected (RS)', colour='rgb(180,0,0)', dash="dash", var="IRS"),
#     str(PARAMS.ISR_ind): dict(name='Infected (SR)', colour='rgb(240,0,0)', dash="dashdot", var="ISR"),
#     str(PARAMS.ISS_ind): dict(name='Infected (SS)', colour='rgb(240,0,0)', dash="solid", var="IS"),
    
#     str(PARAMS.R_ind): dict(name='Removed', colour='black', dash="solid", var="R"),
    
#     str(PARAMS.PRR_ind): dict(name='Primary inoc. (RR)', colour='rgb(150,0,150)', dash="dot", var="PR"),
#     str(PARAMS.PRS_ind): dict(name='Primary inoc. (RS)', colour='rgb(180,0,180)', dash="dash", var="PRS"),
#     str(PARAMS.PSR_ind): dict(name='Primary inoc. (SR)', colour='rgb(210,0,210)', dash="dashdot", var="PSR"),
#     str(PARAMS.PSS_ind): dict(name='Primary inoc. (SS)', colour='rgb(240,0,240)', dash="solid", var="PS"),
    
#     str(PARAMS.fung_1_ind): dict(name='Fung. 1 conc.', colour='rgb(120,240,0)', dash="solid", var="Fung1"),
#     str(PARAMS.fung_2_ind): dict(name='Fung. 2 conc.', colour='rgb(240,120,0)', dash="solid", var="Fung2"),
#     }

E_cols = pltly_clrs.n_colors("rgb(0,0,255)", "rgb(255,255,255)", 5, colortype="rgb")[:4]
        
I_cols = pltly_clrs.n_colors("rgb(255,0,0)", "rgb(255,255,255)", 5, colortype="rgb")[:4]


ATTRS_DICT = {
    'S': dict(color='limegreen', dash='solid', name='Susceptible'),
    'R': dict(color='rgb(100,100,100)', dash='solid', name='Removed'),

    'ERR': dict(color=E_cols[0], dash='dot', name='E (rr)'),
    'ERS': dict(color=E_cols[1], dash='dash', name='E (rs)'),
    'ESR': dict(color=E_cols[2], dash='dashdot', name='E (sr)'),
    'ESS': dict(color=E_cols[3], dash='solid', name='E (ss)'),

    'IRR': dict(color=I_cols[0], dash='dot', name='I (rr)'),
    'IRS': dict(color=I_cols[1], dash='dash', name='I (rs)'),
    'ISR': dict(color=I_cols[2], dash='dashdot', name='I (sr)'),
    'ISS': dict(color=I_cols[3], dash='solid', name='I (ss)'),

    'fung_1': dict(color='turquoise', dash='solid', name='Fungicide A'),
    'fung_2': dict(color='magenta', dash='dot', name='Fungicide B'),
}


LABEL_COLOR = "rgb(110,110,110)"

NULL_HEATMAP_COLOUR = "rgb(100, 100, 100)"

STRAIN_ATTRS = dict(SS=dict(color="rgb(34,140,34)", dash="dot", longname='Double sensitive', abbrv="SS"),
                RS=dict(color="rgb(20,20,200)", dash="dash", longname='Single resistant (RS)', abbrv="RS"),
                SR=dict(color="rgb(200,20,20)", dash="dashdot", longname='Single resistant (SR)', abbrv="SR"),
                RR=dict(color="rgb(50,50,50)", dash="solid", longname='Double resistant', abbrv="RR"),
                )

TITLE_MAP = dict(
    LTY = "Lifetime yield",
    TY = "Total yield",
    FY = "Effective life",
    econ = "Economic life",
    )


# default is half page
PLOT_WIDTH = 600
PLOT_HEIGHT = 400

FULL_PAGE_WIDTH = 800


from tqdm import tqdm
import numpy as np

from .calc_sing_ps_run import SinglePSRun, ScanOutput







class ParameterScan:
    """
    Inputs:
    - config: param scan config detailing runs to do and bounds on params
    - seed: random seed for these runs

    Outputs:
    - ScanOutput object
    """
    def __init__(self, config, seed) -> None:
        self.config = config
        self.seed = seed
        

    def run(self):
        """
        Run random scan over uniform dists and save output
        """

        output = self._get_scan_output()
        
        output.save(self.config, self.seed)




    def _get_scan_output(self):

        np.random.seed(self.seed)
        
        scan_output = ScanOutput()

        N_ITS = self.config["NIts"]

        for r_ind in tqdm(range(N_ITS)):

            run_index = self.seed*N_ITS + r_ind

            this_run = SinglePSRun(self.config, run_index)

            scan_output.add_new_output(this_run.output)

        return scan_output
    



import pandas as pd
import numpy as np






class OtherStratsDF:
    def __init__(self, grid_output) -> None:
        FYs = grid_output.FY
        
        self.df = self._get_strategy_outcomes(FYs)
    


    def _get_strategy_outcomes(self, FYs):
        
        minEqDoseELVec = [float(FYs[i, i]) for i in range(FYs.shape[0])]
        
        data = dict(
                max_grid_EL = np.amax(FYs),
                O_minEqDoseEL = self._get_first_geq_2_element(minEqDoseELVec),
                O_corner_00 = FYs[0, 0],
                O_corner_10 = FYs[-1, 0],
                O_corner_01 = FYs[0, -1],
                O_fullDoseEL = FYs[-1, -1])

        out = pd.DataFrame([data])

        return out
    
    
    
    @staticmethod
    def _get_first_geq_2_element(vec):
        filtered = list(filter(lambda x: x>=2, vec))

        if not filtered:
            return "NA"

        return filtered[0]







import pandas as pd

from model.simulator import RunGrid

from .pars import RandomPars
from .calc_method_contour import RunAlongContourDFs
from .calc_method_IVT import CheckStrategyUsingIVT_DF
from .calc_other_strats import OtherStratsDF






class SinglePSRun:
    def __init__(self, config, run_index) -> None:
        
        self.config = config
        
        self.run_index = run_index

        rand_pars = self._get_rand_pars_obj()
        
        grid_output = RunGrid(rand_pars.fung_parms).run(rand_pars.grid_conf)
        
        self.output = self._get_output(rand_pars, grid_output)



    def _get_output(self, rand_pars, grid_output):
        
        par_df = rand_pars.par_df
        
        RFB_dfs = RunAlongContourDFs(rand_pars, grid_output,
                        self.config['n_cont_points'], "RFB")

        EqSel_dfs = RunAlongContourDFs(rand_pars, grid_output,
                        self.config['n_cont_points'], "EqSel")

        other_strats_df = OtherStratsDF(grid_output).df

        IVT_RFB_df = CheckStrategyUsingIVT_DF(grid_output, "RFB").df

        IVT_EqSel_df = CheckStrategyUsingIVT_DF(grid_output, "EqSel").df

        summary_df = pd.concat([ 
                    RFB_dfs.summary, 
                    EqSel_dfs.summary,
                    other_strats_df,
                    IVT_RFB_df,
                    IVT_EqSel_df],
                    axis=1)


        summary_df['run'] = self.run_index
        RFB_dfs.df['run'] = self.run_index
        EqSel_dfs.df['run'] = self.run_index

        out = ScanOutput()
        out.add_data(par_df, summary_df, RFB_dfs.df, EqSel_dfs.df)

        return out
        



    def _get_rand_pars_obj(self):
    
        RP = RandomPars(self.config, self.run_index)

        RP.find_pars()
        
        RP.get_all_parms_df()

        RP.get_grid_conf(self.config["grid_number"])

        return RP









class ScanOutput:
    """
    Output for one or many PS runs.
    """

    def __init__(self) -> None:
        self.par_df = pd.DataFrame()
        self.summary_df = pd.DataFrame()
        self.RFB_df = pd.DataFrame()
        self.EqSel_df = pd.DataFrame()


    def add_data(self, par_df, summary_df, RFB_df, EqSel_df):
        self.par_df = par_df
        self.summary_df = summary_df
        self.RFB_df = RFB_df
        self.EqSel_df = EqSel_df


    def add_new_output(self, scan_output):
        self.par_df = pd.concat([self.par_df, scan_output.par_df], axis=0)
        self.summary_df = pd.concat([self.summary_df, scan_output.summary_df], axis=0)
        self.RFB_df = pd.concat([self.RFB_df, scan_output.RFB_df], axis=0)
        self.EqSel_df = pd.concat([self.EqSel_df, scan_output.EqSel_df], axis=0)


    def save(self, config, seed):
        
        par_str = config['par_str']

        for key, df in vars(self).items():

            filename = (config['folder_save'] + "/par_scan/" +
                                f"{key}_seed={seed}_{par_str}.csv")
            
            print(f"\n Random Scan {key}, saved as:\n {filename}")

            df_out = self._move_run_to_first_col(df)

            df_out.to_csv(filename, index=False)




    def _move_run_to_first_col(self, df):
        cols = list(df.columns)
        new_cols = ['run'] + [cc for cc in cols if cc!='run']
        out = df.loc[:, new_cols]
        return out



import itertools
import numpy as np
from scipy.optimize import minimize


from model.utils import logit10_difference
from model.simulator import RunSingleTactic


# RFB Finder

def find_RFB_from_single_run(single_run):
    
    fy = single_run.failure_year

    end_freqs = single_run.end_freqs

    ef1 = end_freqs['RS']
    ef2 = end_freqs['SR']

    out = get_RFB(ef1, ef2, fy)
    return out




def get_RFB(end_freqs_1, end_freqs_2, fy):
    fy = int(fy)

    if not fy>1:
        return None
    
    r1 = end_freqs_1[fy-1]
    r2 = end_freqs_2[fy-1]

    try:
        return logit10_difference(r1, r2)
    except Exception as e:
        print(f"find_RFB warning/error: {e} \n rfs={r1, r2}")
        return None





# FY Sel Finder

def find_FY_sel_from_single_run(single_run):
    fy = single_run.failure_year
    
    start_freqs = single_run.start_freqs
    end_freqs = single_run.end_freqs

    ef1 = end_freqs['RS'][0]
    sf1 = start_freqs['RS'][0]
    
    ef2 = end_freqs['SR'][0]
    sf2 = start_freqs['SR'][0]

    out = get_FY_sel(ef1, sf1, ef2, sf2, fy)
    return out
    
    


def get_FY_sel(ef1, sf1, ef2, sf2, fy):
    """
    NB have changed so it compares end of season and start of season,
    not start of consecutive seasons. 
    
    This is because sexual reproduction can cause a change that wasn't due 
    to the tactic but was just down to ratios starting away from those expected
    in a perfectly mixed population.
    """
    
    fy = int(fy)

    if not fy>1:
        return None
    
    sr1 = ef1/sf1
    sr2 = ef2/sf2
    
    try:
        return sr1/(sr1+sr2)
    except Exception as e:
        print(f"find_FY_sel warning/error: {e} \n srs={(sr1,sr2)}")
        return None











class StrategyArray:
    def __init__(self, grid_output) -> None:
        self.grid_output = grid_output        
        self.contour_finder = ContourDoseFinder



    def _check_valid(self):
        """
        Check if Strategy is a possible tactic:
            - does the contour exist for this array?
        """
        return (np.nanmax(self.array)>self.level
                            and np.nanmin(self.array)<self.level)


    def find_contours(self, rand_pars, n_cont_points):

        if (not self.is_valid):
            return {}

        DS_extremes = DoseSumExtremes(self.array, self.level)

        if ((DS_extremes.min is None) or 
                (DS_extremes.max is None)):
            return {}
        

        cf = self.contour_finder(rand_pars,
                            self.name,
                            DS_extremes, 
                            n_cont_points,
                            self.level)
        
        contours = cf.get_doses_on_contour()
        
        return contours






class EqualResFreqBreakdownArray(StrategyArray):
    def __init__(self, grid_output) -> None:
        super().__init__(grid_output)

        self.level = 0
        self.name = "RFB"

        self.array = self._generate_array()
        self.is_valid = self._check_valid()
        
    
        
    def _generate_array(self):

        FYs = self.grid_output.FY
        end_freqs = self.grid_output.end_freqs_DA

        out = np.ones(FYs.shape)
        
        for i, j in itertools.product(range(out.shape[0]), 
                                        range(out.shape[1])):
            
            fy = FYs[i,j]
                
            ef1 = end_freqs['RS'][i,j,:]
            ef2 = end_freqs['SR'][i,j,:]

            out[i,j] = get_RFB(ef1, ef2, fy)

        return out




# * End of ERFB cls









class EqualSelectionArray(StrategyArray):
    def __init__(self, grid_output) -> None:
        super().__init__(grid_output)

        self.level = 0.5
        self.name = "EqSel"        

        self.array = self._generate_array()
        self.is_valid = self._check_valid()



    def _generate_array(self):

        FYs = self.grid_output.FY
        start_freqs = self.grid_output.start_freqs_DA
        end_freqs = self.grid_output.end_freqs_DA
        
        out = np.ones(FYs.shape)
        
        for i, j in itertools.product(range(out.shape[0]), 
                                        range(out.shape[1])):
            
            fy = FYs[i,j]

            ef1 = end_freqs['RS'][i,j,0]
            sf1 = start_freqs['RS'][i,j,0]
            
            ef2 = end_freqs['SR'][i,j,0]
            sf2 = start_freqs['SR'][i,j,0]

            out[i,j] = get_FY_sel(ef1, sf1, ef2, sf2, fy)

        return out





# * End of ES cls











class ContourDoseFinder:
    def __init__(self, rand_pars, 
                        strat_name,
                        DS_extremes,
                        n_cont_points,
                        level,
                        tol=0.001) -> None:

        self.rand_pars = rand_pars
        
        if strat_name=="RFB":
            self.cont_quant_finder = find_RFB_from_single_run
            
        elif strat_name=="EqSel":
            self.cont_quant_finder = find_FY_sel_from_single_run
            
        else:
            raise Exception(f"invalid strat_name: {strat_name}")

        self.DS_extremes = DS_extremes
        self.n_cont_points = n_cont_points
        self.level = level

        self.CONT_DIST_THRESH = tol



    def get_doses_on_contour(self):
        
        DS_bds = self.DS_extremes

        dose_sums = np.linspace(DS_bds.min, DS_bds.max, self.n_cont_points)

        x_list = []
        y_list = []
        cont_vals = []

        for ds in dose_sums:
            self.dose_sum = ds

            doses = self._get_doses_on_cntr_this_DS(ds)

            if doses is None:
                continue

            
            if self.model_cont_quant is None:
                continue

            # only add if have got close to the contour 
            # ? - why doesn't it get close otherwise?

            dist_from_contour = abs(self.model_cont_quant-self.level)
        
            if dist_from_contour < self.CONT_DIST_THRESH:
                x_list.append(doses['x'])
                y_list.append(doses['y'])
                cont_vals.append(self.model_cont_quant)
            else:
                print("\n")
                print("this run didn't get close to the contour?? ...")
                print("contour level:", self.model_cont_quant)
                print("dose sum:", self.dose_sum)
                print(DS_bds)


        return dict(x=x_list, y=y_list, cont_vals=cont_vals)




    
    def _get_doses_on_cntr_this_DS(self, ds):
        
        lower, upper = self._get_bnds(ds)
        
        x0 = [0.5*(lower+upper)]
        
        bnds = ((lower, upper), )

        try:
            thisFit = minimize(self.objective_fn, x0, bounds=bnds)
            k = thisFit.x[0]
            doses = dict(
                x = 0.5*ds - k,
                y = 0.5*ds + k,
                )
            return doses
        except Exception as e:
            print(e)
            return None
        




    def _get_bnds(self, ds):
        d = 0.5*ds

        # pick lower/upper so that each dose in [0,1]
        lower = max(-d, d - 1)
        upper = min( d,  1 - d)

        return lower, upper






    def objective_fn(self, param):

        k = param[0]

        dose1 = 0.5*self.dose_sum - k
        dose2 = 0.5*self.dose_sum + k

        rp = self.rand_pars

        this_dose_conf = rp.get_single_conf(dose1, dose2)
        this_dose_conf.load_saved = False

        sing_run = RunSingleTactic(rp.fung_parms).run(this_dose_conf)
        
        self.model_cont_quant = self.cont_quant_finder(sing_run)

        dist = self._get_distance_from_contour(k)

        return dist




    def _get_distance_from_contour(self, k):
        """
        If quantity not defined (e.g. this dose combo is in unacceptable region)
        then return a high distance and one which pushes 
        the optimiser back towards the middle of the x+y=DS line
        """
        if self.model_cont_quant is None:
            return abs(k) + 5
        else:
            return (self.model_cont_quant - self.level)**2




                














class DoseSumExtremes:

    def __init__(self, z, level) -> None:
        self.min = None
        self.max = None
        
        self.get_min_and_max_valid_DS(z, level)



    def get_min_and_max_valid_DS(self, z, level):
        """
        Check for each dose sum whether there are values above and below 'level'
        """
        # relative to 0
        zz = np.array(z) - level

        ds_vec = np.linspace(0, 2, -1+2*z.shape[0])

        valid_ds_list = []

        for ds_ind in range(len(ds_vec)):
            is_valid = self._check_this_ds_straddles_level(zz, ds_ind, z.shape[0])
            
            if is_valid:
                valid_ds_list.append(ds_vec[ds_ind])


        if not len(valid_ds_list):
            return None

        self.min = min(valid_ds_list)
        self.max = max(valid_ds_list)






    def _check_this_ds_straddles_level(self, zz, ds_ind, n):
        vals = []

        bottom = max(0, 1 + ds_ind-n)
        top = min(ds_ind, n-1)
        
        for ii in range(bottom, 1+top):
            if not np.isnan(zz[ii, ds_ind-ii]):
                vals.append(zz[ii, ds_ind-ii])

        if not len(vals):
            return False

        if min(vals)<0 and max(vals)>0:
            return True
        
        return False









"""
Analyse the parameter scan output
"""

from ..fns.config import config_rand
from ..fns.post_process import PostProcess


def main(config):
    
    PP = PostProcess(config['folder_save'], config['par_str'])

    PP.get_maximum_along_contour_df()
    
    PP.analyse_max_contour_df()

    PP.analyse_failed()
    
    # PP.which_runs_worked_max_cont()

    # PP.re_run(NDoses=6, run_indices=list(range(1)))
    # PP.re_run(NDoses=41, run_indices=[706, 538, 548, 403, 490, 474, 367, 34])
    PP.re_run(NDoses=101, run_indices=[538])

    PP.check_high_or_low_dose()



if __name__=="__main__":
    main(config_rand)


"""
Scan over: 
- IRFs
- fung asymptote
- fung decay rate
- partial res
- levels of SR

Process
- run on cluster,
- one seed for each chunk, 
- compile into one df using r2_combine_outputs.py
"""

import sys

from ..fns.calc_full_scan import ParameterScan
from ..fns.config import config_rand


def main(config, seed):
    ParameterScan(config, seed).run()


if __name__=="__main__":
    
    if len(sys.argv)!=2:
        raise Exception("Supply one argument: a random seed")

    seed = int(sys.argv[1])
    main(config_rand, seed)


from model.simulator import RunSingleTactic

from model.config import ConfigSingleRun

from model.config_classes import SingleConfig

from plotting.figures import yield_by_year, res_freqs_single_t_plot, \
    single_year_plot, yield_res_freqs_plot, plot_frequencies, plot_frequencies_over_time


# which plots

yield_single = False
res_freqs_single = False
yield_res_freqs = False
single_year = True
freq_bar_plot = False
freq_time_plot = False

# run
bools = [res_freqs_single, freq_bar_plot]

if any(bools):
    output = RunSingleTactic().run(ConfigSingleRun)


# plots
conf_str = ConfigSingleRun.config_string_img

if yield_single:
    rf1 = 10**(-4)
    rf2 = 10**(-2)
    ConfigYRF = SingleConfig(20, rf1, rf2, 1, 1, 1, 1)

    output = RunSingleTactic().run(ConfigYRF)

    yield_by_year(output, conf_str)


if yield_res_freqs:
    rf1 = 10**(-4)
    rf2 = 10**(-2)
    ConfigYRF = SingleConfig(20, rf1, rf2, 1, 1, 1, 1)

    output = RunSingleTactic().run(ConfigYRF)

    yield_res_freqs_plot(output, ConfigYRF.config_string_img)
    

if res_freqs_single:
    res_freqs_single_t_plot(output, conf_str)


if single_year:
    rf1 = 10**(-1)
    rf2 = 2*10**(-1)
    ConfigSY = SingleConfig(20, rf1, rf2, 1, 1, 0.5, 0.5)
    output = RunSingleTactic().run(ConfigSY)

    indices_use = dict(
        # all = list(range(16)),
        sus_rec = [0, 9],
        infected = list(range(5,9)),
        fungs = [14,15],
        )
    
    for key in indices_use:
        indices_to_plot = indices_use[key]
        single_year_plot(output, indices_to_plot, conf_str)


if freq_bar_plot:
    plot_frequencies(output, conf_str)
    

if freq_time_plot:
    rf1 = 10**(-4)
    rf2 = 10**(-2)
    ConfigFTP = SingleConfig(10, rf1, rf2, 1, 1, 1, 1)
    ConfigFTP.sex_prop = 0
    ConfigFTP.add_string()

    output = RunSingleTactic().run(ConfigFTP)

    plot_frequencies_over_time(output, ConfigFTP.config_string_img)




# proper/fast/prev_run
run_type = "proper"

baseline_dec = 1.11*10**(-2)

baseline = {
        "RFS1": [-8,-2],
        "RFS2": [-8,-2],
        "RFD": [-15,-3],

        "asym1": [0.4, 1],
        "asym2": [0.4, 1],
        
        "theta": 12,

        "dec_rate1": [(1/3)*baseline_dec, 3*baseline_dec],
        "dec_rate2": [(1/3)*baseline_dec, 3*baseline_dec],

        "SR": [0,1],
        
        "load_saved": False,
        "save": False,
        "folder_save": "./param_scan/outputs",
        }


if run_type == "proper":
    run_pars = {
        "grid_number": 51,
        "n_cont_points": 71,
        "n_years": 40,
        "NIts": 15,
        }

elif run_type == "fast":
    run_pars = {
        "grid_number": 5,
        "n_cont_points": 3,
        "n_years": 35,
        "NIts": 2,
        }

elif run_type == "prev_run":
    run_pars = {
        "grid_number": 51,
        "n_cont_points": 5,
        "n_years": 35,
        "NIts": 5,   
        }




def get_par_str(config):
    string = ""
    
    for key in config.keys():
        attribute = config[key]

        if key in ["load_saved", "save", "folder_save"]:
            continue

        if type(attribute)==float or type(attribute)==int:
            string += f"{str(key)[:2]}={attribute}"
        else:
            string += f"{str(key)[:2]}=L{attribute[0]},U{attribute[1]}"

    string = string.replace(".", ",")

    return string




config_rand = {**baseline, **run_pars}
config_rand['par_str'] = get_par_str(config_rand)


import pandas as pd
import numpy as np

from model.simulator import RunSingleTactic
from model.strategy_arrays import EqualResFreqBreakdownArray, \
    EqualSelectionArray


# TOC
# RunAlongContourDF

# ThisStratDetailedDF
# ThisStratSummaryDF




class RunAlongContourDFs:
    def __init__(self, rand_pars, grid_output,
                        n_cont_points, strat_name) -> None:
        
        print(f"Running contour method: {strat_name}")

        if strat_name=="RFB":
            strat_obj = EqualResFreqBreakdownArray(grid_output)
        elif strat_name=="EqSel":
            strat_obj = EqualSelectionArray(grid_output)
        else:
            raise Exception(f"invalid strat_name: {strat_name}")

        
        
        max_grid_EL = np.amax(grid_output.FY)

        self.df = ThisStratDetailedDF(rand_pars, 
                                    n_cont_points, 
                                    strat_obj,
                                    max_grid_EL).df
        
        
        self.summary = ThisStratSummaryDF(self.df, 
                                        strat_obj.name,
                                        strat_obj.level,
                                        max_grid_EL).df




# End of RunAlongContourDF






class ThisStratDetailedDF:
    def __init__(self, rand_pars, n_cont_points, strat_obj, max_grid_EL) -> None:
        
        self.rand_pars = rand_pars
        self.n_cont_points = n_cont_points

        self.strat_name = strat_obj.name
        self.level = strat_obj.level

        self.strat_obj = strat_obj
        self.max_grid_EL = max_grid_EL

        self.df = self._get_df()
        


    def _get_df(self):
        
        cntr = self.strat_obj.find_contours(self.rand_pars,
                                            self.n_cont_points)
        out = self._find_df(cntr)
        

        return out






    def _find_df(self, contour):
        
        df_this_run = pd.DataFrame()

        if not contour:
            return df_this_run


        for dose1, dose2, cont_quant in zip(contour['x'], contour['y'], contour['cont_vals']): 

            EL = self._get_EL_this_dose_combo(dose1, dose2)
            
            data = {"cont_quant": cont_quant,
                     "dose1": dose1,
                     "dose2": dose2,
                     "DS": dose1 + dose2,
                     "EL": EL}

            df_this_run = df_this_run.append(data, ignore_index=True)
        
        
        df_this_run['worked'] = max(df_this_run['EL'])>=self.max_grid_EL
        df_this_run['max_grid_EL'] = self.max_grid_EL

        return df_this_run





    def _get_EL_this_dose_combo(self, dose1, dose2):

        rp = self.rand_pars

        this_dose_conf = rp.get_single_conf(dose1, dose2)
                
        sing_run = RunSingleTactic(rp.fung_parms).run(this_dose_conf)
        
        EL = sing_run.failure_year
        
        return EL




# End of ThisStratDetailedDF












class ThisStratSummaryDF:

    def __init__(self, df, strat_name, level, max_grid_EL) -> None:
        self.df = df
        self.strat_name = strat_name
        self.level = level
        self.max_grid_EL = max_grid_EL

        self.LOW_THRESH = 1/3
        self.HIGH_THRESH = 2/3

        self.df = self.get_summary_df()





    def get_summary_df(self):
        df = self.df
        strat_name = self.strat_name

        if not df.shape[0]:
            return pd.DataFrame()

        lowDS_val = self._get_low_dose_max_EL()
        medDS_val = self._get_med_dose_max_EL()
        highDS_val = self._get_high_dose_max_EL()

        
        min_opt_dist_from_cntr = self._get_min_opt_dist_from_contour()
        

        data= {
            f"c_{strat_name[0]}_minContEL": min(df['EL']),
            f"c_{strat_name[0]}_maxContEL": max(df['EL']),

            f"c_{strat_name[0]}_min_opt_dist_from_contour": min_opt_dist_from_cntr,

            f"c_{strat_name[0]}_minDS": min(df['DS']),
            f"c_{strat_name[0]}_maxDS": max(df['DS']),
            
            f"c_{strat_name[0]}_worked_geq": self.max_grid_EL<=max(df['EL']),
            f"c_{strat_name[0]}_worked_equal": self.max_grid_EL==max(df['EL']),
            
            f"c_{strat_name[0]}_lowDoseMaxEL": lowDS_val,
            f"c_{strat_name[0]}_medDoseMaxEL": medDS_val,
            f"c_{strat_name[0]}_highDoseMaxEL": highDS_val,
            }

        return pd.DataFrame([data])


    


    
    
    def _get_low_dose_max_EL(self):
        df = self.df

        DS_thres_low = min(df['DS']) + self.LOW_THRESH*(max(df['DS']) - min(df['DS']))
        
        filt = df[df['DS'] < DS_thres_low]

        return self._get_max_if_df_non_empty(filt)
    





    def _get_med_dose_max_EL(self):
        df = self.df

        DS_thres_low = min(df['DS']) + self.LOW_THRESH*(max(df['DS']) - min(df['DS']))
        DS_thres_high = min(df['DS']) + self.HIGH_THRESH*(max(df['DS']) - min(df['DS']))

        filt = df[((df['DS'] >= DS_thres_low) & (df['DS'] <= DS_thres_high))]

        return self._get_max_if_df_non_empty(filt)

    




    def _get_high_dose_max_EL(self):
        df = self.df

        DS_thres_high = min(df['DS']) + self.HIGH_THRESH*(max(df['DS']) - min(df['DS']))

        filt = df[df['DS'] > DS_thres_high]
        
        return self._get_max_if_df_non_empty(filt)
        



    def _get_max_if_df_non_empty(self, df):
        if df.shape[0]:
            return max(df['EL'])
        else:
            return "NA"




    def _get_min_opt_dist_from_contour(self):
        df = self.df

        opt_df = df[df['EL']==self.max_grid_EL]

        vec = abs(opt_df["cont_quant"] - self.level)

        if len(vec):
            out = min(vec)
            return out
        else:
            return "NA"
        




# End of ThisStratSummaryDF





    







from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()



setup(name='hrhr',
      version='0.0.3',
      description='HRHR model',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/nt409/HRHR',
      author='Nick Taylor',
      author_email='nt409@cam.ac.uk',
      license='MIT',
      packages=['model'],
    #   packages=find_packages(),
      zip_safe=False)


import numpy as np
from math import exp, log10
from scipy.integrate import simps
import os
import pickle


from model.params import PARAMS


# * Utility functions

def object_dump(file_name, object_to_dump):
    
    # check if file path exists - if not create
    outdir =  os.path.dirname(file_name)
    if not os.path.exists(outdir):
        os.makedirs(outdir,exist_ok=True) 
        
    with open(file_name, 'wb') as handle:
        pickle.dump(object_to_dump, handle, protocol=pickle.HIGHEST_PROTOCOL) # protocol?

def logit10(x):
    if x>0 and x<1:
        return log10(x/(1-x))
    else:
        raise Exception(f"x={x} - invalid value")


def logit10_difference(x1, x2):
    return logit10(x1) - logit10(x2)

def log10_difference(x1, x2):
    return log10(x1) - log10(x2)

# * End of utility functions




# * Simulatr functions

def yield_calculator(y, t):
    out = simps(y, t)
    return out




def res_prop_calculator(solution):
    """
    Uses final value of disease densities (end of season) to determine the res props.

    These are used for next season (with a SR step in between if sr_prop=/=0)
    """

    disease = (solution.IRR[-1] + 
                    solution.IRS[-1] +
                    solution.ISR[-1] + 
                    solution.ISS[-1])
        
    Res_disease_1 = solution.IRR[-1] + solution.IRS[-1]
    Res_disease_2 = solution.IRR[-1] + solution.ISR[-1]

    res_props_out = dict(
        f1 = Res_disease_1/disease,
        f2 = Res_disease_2/disease,
        )
    
    strain_frequencies = dict(
        RR = solution.IRR[-1]/disease,
        RS = solution.IRS[-1]/disease,
        SR = solution.ISR[-1]/disease,
        SS = solution.ISS[-1]/disease
        )
    
    return res_props_out, strain_frequencies


# * End of Simulatr functions





    
# * Classes

class Fungicide:
    def __init__(self, omega, theta, delta):
        self.omega = omega
        self.theta = theta
        self.delta = delta

    def effect(self, conc):
        effect = 1 - self.omega*(1 - exp(- self.theta * conc))
        return effect

# * End of Fcide cls










class FungicideStrategy:
    def __init__(self, my_strategy, n_seasons):
        self.my_strategy = my_strategy
        self.n_seasons = n_seasons



    def get_grid_doses(self, f1_val, f2_val, n_doses):

        self.conc_f1 = f1_val/(n_doses-1)
        self.conc_f2 = f2_val/(n_doses-1)

        self._get_doses_for_this_strategy()      

        return self.fung1_doses, self.fung2_doses



    def _get_doses_for_this_strategy(self):
        if self.my_strategy=='mix':
            self._get_mixed_doses()

        elif self.my_strategy=='alt_12':
            self._get_alt_12_doses()

        elif self.my_strategy=='alt_21':
            self._get_alt_21_doses()

        else:
            raise Exception(f"Invalid strategy named: {self.my_strategy}")




    def _get_mixed_doses(self):        
        # did half 0.5*
        # but Hobbelen paper just says it means twice as much
        self.fung1_doses = dict(
            spray_1 = self.conc_f1*np.ones(self.n_seasons),
            spray_2 = self.conc_f1*np.ones(self.n_seasons)
            )
        self.fung2_doses = dict(
            spray_1 = self.conc_f2*np.ones(self.n_seasons),
            spray_2 = self.conc_f2*np.ones(self.n_seasons)
            )


    def _get_alt_12_doses(self):
        self.fung1_doses = dict(
            spray_1 = self.conc_f1*np.ones(self.n_seasons),
            spray_2 = np.zeros(self.n_seasons)
            )
        self.fung2_doses = dict(
            spray_1 = np.zeros(self.n_seasons),
            spray_2 = self.conc_f2*np.ones(self.n_seasons)
            )
    

    def _get_alt_21_doses(self):
        self.fung1_doses = dict(
            spray_1 = np.zeros(self.n_seasons),
            spray_2 = self.conc_f1*np.ones(self.n_seasons)
            )
        self.fung2_doses = dict(
            spray_1 = self.conc_f2*np.ones(self.n_seasons),
            spray_2 = np.zeros(self.n_seasons)
            )

# * End of FcideStrt cls








class SelectionFinder:    
    def __init__(self, primary_inoc, final_res_dict) -> None:
        self._get_init_res_dict(primary_inoc)

        self.final_res_dict = final_res_dict

        self._get_selection()


    def _get_init_res_dict(self, primary_inoc):
        self.initial_res_dict = dict(
            f1 = primary_inoc['RR'] + primary_inoc['RS'],
            f2 = primary_inoc['RR'] + primary_inoc['SR']
            ) 


    def _get_selection(self):
        self.sel = dict(f1=1, f2=1)

        for key in ['f1','f2']:
            self._get_sel_this_fung(key)
        


    def _get_sel_this_fung(self, key):
        in_res_dict = self.initial_res_dict
        fn_res_dict = self.final_res_dict
        
        if in_res_dict[key] > 0:
            self.sel[key] = fn_res_dict[key] / (in_res_dict[key]/PARAMS.init_den)

# * End of SelFinder cls




